/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "base.h"

#include <cerrno>
#include <cstring>

#if IVALUE==IINT
#include <climits>
#endif

using namespace std;

#ifndef HAVE_EXP10_SET
ieee exp10(ieee x){
  return exp(LOG10*x);
}
#endif

/*
value::value(){
  type=INTEGER;
}
*/

/*
value::value(itype ii){
  type=INTEGER;
  i=ii;
  d=ii;
}
*/

/*
value::value(ieee ff){
  type=FLOATING;
  d=ff;
}
*/

itype value::intpart() const{
  if(d>=0){
    return (itype)d;
  }else{
    return (itype)(d-(itype)d+1)+(itype)d-1;
  }
}

value::value(const char* s){
  if(strchr(s,'.')||strchr(s,'e')||strchr(s,'E')){
    type=FLOATING;
#if VALUE==LONG
    d=strtold(s,0);
#elif VALUE==DOUBLE
    d=strtod(s,0);
#else
    d=strtof(s,0);
#endif
    return;
  }else{
    type=INTEGER;
#if IVALUE==ILONG
    errno=0;
    long long int ival=strtoll(s,0,10);
    if(errno){
      fprintf(stderr,"parse: integer overflow during parsing input\n");
      type=FLOATING;
      d=NAN;
      i=-1;
    }
    else{
      i=ival;
      d=i;
    }

#else
    errno=0;
    long int ival=strtol(s,0,10);
    if(errno || ival>INT_MAX || ival<INT_MIN){
      fprintf(stderr,"parse: integer overflow during parsing input\n");
      type=FLOATING;
      d=NAN;
      i=-1;
    }else{
      i=ival;
      d=i;
    }

#endif
  }
}

itype value::assertI(const char* op){
  if(type!=INTEGER){
    fprintf(stderr,"warning: integer required by '%s', converting\n",op);
    return intpart();
  }
  return i;
}

ieee value::assertD(const char* op){
  if(type!=FLOATING){
    fprintf(stderr,"warning: floating required by '%s', converting\n",op);
  }
  return d;
}

value value::operator+(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      return value(d+v.d);
    }else{
      return value(result);
    }

#else
    return value(i+v.i);
#endif
  }else{
    return value(d+v.d);
  }
}

value value::operator-(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      return value(d-v.d);
    }else{
      return value(result);
    }

#else
    return value(i-v.i);
#endif
  }else{
    return value(d-v.d);
  }
}


value value::operator*(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      return value(d*v.d);
    }else{
      return value(result);
    }

#else
    return value(i*v.i);
#endif
  }else{
    return value(d*v.d);
  }
}

value value::operator/(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    if(v.i==0)return value((itype)0);
    else return value(i/v.i);
  }else{
    if(v.d==0)return value((itype)0);
    else return value(d/v.d);
  }
}

static value flpow(ieee i1, itype i2){
  if(i2<0){
    return value(pow(i1,(int)i2));
  }else{
    int sign=(i1<0) && (i2&1);
    i1=i1<0?-i1:i1;
    ieee trailing=i1;
    ieee total=1;
    if(i2&1)total*=trailing;
    i2>>=1;
    while(i2!=0){
      trailing*=trailing;
      if(i2&1)total*=trailing;
      i2>>=1;
    }
    if(sign)total=-total;
    return value(total);
  }
}

inline static itype imul(itype a, itype b, char& flag){
#ifdef TRAP_OVR
  imul_(a,b,flag);
#if DEBUG>0
  if(flag){
    fprintf(stderr,"runtime: integer overflow on multiplication\n");
  }
#endif
  return a;  
#else
  return a*b;
#endif
}


static value intpow(itype i1, itype i2){
  if(i2<0){
    return value(pow((ieee)i1,(int)i2));
  }else{
    char flag=0;
    int sign=(i1<0) && (i2&1);
    itype base=i1<0?-i1:i1;
    itype trailing=base;
    itype total=1;
    itype ex=i2;
    if(ex&1)total=imul(total,trailing,flag);
    ex>>=1;
    while(ex!=0){
      trailing=imul(trailing,trailing,flag);
      if(ex&1)total=imul(total,trailing,flag);
      ex>>=1;
    }
    if(flag)return flpow((ieee)i1,i2);
    if(sign)total=-total;
    return value((itype)total);
  }
}

value value::operator^(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    return intpow(i,v.i);
  }else if(v.type==INTEGER){
    return flpow(d,v.i);
  }else{
    return value(pow(d,v.d));
  }
}

static ieee d_mod(ieee d1, ieee d2){
  ieee r=d1-(floor(d1/d2))*d2;
  return r<0?0:r;
}

static ieee f_mod(ieee v1, ieee v2){
  if(v2==0)return v1;
  if(v2>=0){
    return d_mod(v1,v2);
  }else{
    v2=-v2;
    return d_mod(v1,v2)-v2;
  }
}

static itype i_mod(itype i1, itype i2){
  if(i2==0)return i1;
  if(i2>=0){
    //return i1<0?(i2-(-i1)%i2)%i2:i1%i2;//room for improvement
    if(i1<0){
      itype it=(i2-(-i1)%i2);
      if(it==i2)return 0;
      else return it;
    }else{
      return i1%i2;
    }
  }else{
    i2=-i2;
    itype it;
    if(i1<0){
      it=(i2-(-i1)%i2);
      if(it==i2)it=0;
    }else{
      it=i1%i2;
    }
    return it-i2;
    //return (i1<0?(i2-(-i1)%i2)%i2:i1%i2)-i2;
  }
}

value value::operator%(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    return value(i_mod(i,v.i));
  }else{
    return value(f_mod(d,v.d));
  }
}

value value::e(const value& v) const {
  return value(d*exp10(v.d));
}

value value::binomial(const value& v) const {
  if(v.type==INTEGER){
    value total=(itype)1;
    value nom=(*this);
    for(int i=1;i<=v.i;i++){
      total*=nom;
      total=total/(itype)i;
      nom=nom-(itype)1;
    }
    return total;
  }else{
    int i1, i2, i3;
    ieee g1=lgamma_r(d+1,&i1);
    ieee g2=lgamma_r(v.d+1,&i2);
    ieee g3=lgamma_r(d-v.d+1,&i3);
    return value((ieee)((i1*i2*i3)*exp(g1-g2-g3)));
  }
}



void value::operator+=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      type=FLOATING;
      d+=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i+=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d+=v.d;
  }
}

void value::operator-=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      type=FLOATING;
      d-=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i-=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d-=v.d;
  }
}


void value::operator*=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      type=FLOATING;
      d*=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i*=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d*=v.d;
  }
}

void value::operator/=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){
    if(v.i==0){
      i=0;
      d=0;
    }else{
      i/=v.i;
      d=i;
    }
  }else{
    if(v.d==0){
      type=INTEGER;
      i=0;
      d=0;      
    }else{
      type=FLOATING;
      d/=v.d;
    }
  }
}



value value::operator+(const ieee& v) const{
  return value(d+v);
}

value value::operator-(const ieee& v) const{
  return value(d-v);
}

value value::operator*(const ieee& v) const{
  return value(d*v);
}

value value::operator/(const ieee& v) const{
  if(v==0)return value((itype)0);
  return value(d/v);
}

value value::operator^(const ieee& v) const{
  return value(pow(d,v));
}

value value::operator%(const ieee& v) const{
  return value(f_mod(d,v));
}


value value::operator+(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      return value(d+(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i+v);
#endif
  }else{
    return value(d+v);
  }
}

value value::operator-(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      return value(d-(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i-v);
#endif
  }else{
    return value(d-v);
  }
}

value value::operator*(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;

    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      return value(d*(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i*v);
#endif
  }else{
    return value(d*v);
  }
}

value value::operator/(const itype& v) const{
  if(v==0)return value((itype)0);
  if(type==INTEGER){
    return value(i/v);
  }else{
    return value(d/v);
  }
}

value value::operator^(const itype& v) const{
  if(type==INTEGER){
    return intpow(i,v);
  }else {
    return flpow(d,v);
  }
}

value value::operator%(const itype& v) const{
  if(type==INTEGER){
    return value(i_mod(i,v));
  }else{
    return value(f_mod(d,v));
  }
}

value value::operator-() const{
  if(type==INTEGER){
    return value(-i);
  }else{
    return value(-d);
  }
}

void value::op_neg(){
  i=-i;
  d=-d;
  
}

void value::op_sin(){
  d=sin(d);
  type=FLOATING;
}

void value::op_cos(){
  d=cos(d);
  type=FLOATING;
}

void value::op_tan(){
  d=tan(d);
  type=FLOATING;
}

void value::op_atan(){
  d=atan(d);
  type=FLOATING;
}

void value::op_asin(){
  d=asin(d);
  type=FLOATING;
}

void value::op_acos(){
  d=acos(d);
  type=FLOATING;
}

void value::op_abs(){
  if(type==INTEGER){
    if(i<0){i=-i;d=-d;}
  }else{
    if(d<0)d=-d;
  }
}

void value::op_exp(){
  d=exp(d);
  type=FLOATING;
}

void value::op_sqrt(){
  d=sqrt(d);
  type=FLOATING;
}

void value::op_cbrt(){
  d=cbrt(d);
  type=FLOATING;
}

void value::op_log(){
  d=log(d);
  type=FLOATING;
}


value value::op_bj(int n){
  return value((ieee)jn(n,d));
}

value value::op_by(int n){
  return value((ieee)yn(n,d));
}

void value::op_int(){
  if(type==FLOATING){
    type=INTEGER;
    i=intpart();
    d=i;
  }
}

void value::op_float(){
  type=FLOATING;
}

void value::op_frac(){
  if(type==FLOATING){
   itype ipart=intpart();
   d-=ipart;
  }else{
    type=FLOATING;
    d=0;
  }
}

void value::op_round(){
  if(type==FLOATING){
   i=intpart();
   d-=i;
   if(d>=0.5)i++;
   type=INTEGER;
   d=i;
  }
}


#if IVALUE==ILONG
#define MAX_FACT_INT 20
#else
#define MAX_FACT_INT 12
#endif

void value::op_fact(){
  if(type==INTEGER){
    if(i>MAX_FACT_INT){
      d=1;
      for(itype u=2;u<=i;u++)d*=u;
      type=FLOATING;
    }else{
      itype fac=1;
      for(itype u=2;u<=i;u++)fac*=u;
      d=i=fac;
    }
  }else{
    int gi;
    ieee lg=lgamma_r(d+1,&gi);
    d=gi*exp(lg);
  }
}

void value::op_exp10(){
  d=exp10(d);
  type=FLOATING;
}

void value::op_rand(){
  if(type==INTEGER){
    if(!i)return;
    int sign=i<0?-1:1;
    i=sign*(rand()%(sign*i));
    d=i;
  }else{
    d=(d*rand())/RAND_MAX;
  }
}

int value::is_odd() const{
  if(type!=INTEGER)return 0;
  return i%2!=0;
}

int value::is_even() const{
  if(type!=INTEGER)return 0;
  return i%2==0;
}

int value::is_positive() const{
  if(type==INTEGER)return i>=0;
  else return d>=0;
}

int value::is_negative() const{
  if(type==INTEGER)return i<0;
  else return d<0;
}

int value::is_int() const{
  return type==INTEGER;
}

int value::is_float() const{
  return type==FLOATING;
}



//from Remco's Ramblings
//http://remcobloemen.nl/2011/miller-rabin-primality-test.html (11.08.2012)
//Miller-Rabin test
static bool miller_rabin(uint64_t n, uint64_t k){
  if(n == k) return true;
  uint64_t s, d, b, e, x;

  // Factor n-1 as d 2^s
  for(s = 0, d = n - 1; !(d & 1); s++)d >>= 1;

  // x = k^d mod n using exponentiation by squaring
  for(x = 1, b = k%n, e = d; e; e>>=1){
#ifdef EXTENDED_PRIMES
    if(e&1) x=imulmod_(x,b,n);
    //if(b<65536)b*=b;else
    b = imulmod_(b,b,n);
#else
  // The squaring overflows for n >= 2^32
    if(e&1) x=(x*b)%n;
    b = (b*b)%n;
#endif
  }

  // Verify k^(d 2^[0 s-1]) mod n != 1
  if(x == 1 || x == n-1) return true;
  while(s-- > 1){
#ifdef EXTENDED_PRIMES
    x = imulmod_(x,x,n);
#else
    x = (x*x)%n;
#endif
    if(x == 1) return false;
    if(x == n-1) return true;
  }
  return false;
}


int value::is_prime() const{
  static const int primes[]={3, 5, 7, 11, 13, 17, 19, 23, 29, 
  			     31, 37, 41, 43, 47, 53, 59};
  if(type!=INTEGER)return 0;
  if(i<2)return 0;//negative, 0 or 1
  if(i==2)return 1;//special case
  if(!(i&1))return 0;//number two is special
  //if it is a small prime, it is probably better detected by miller-rabin
  if(i>primes[sizeof(primes)/sizeof(int)-1]){
    for(int a=0;a<sizeof(primes)/sizeof(int);a++){
      if(!(i%primes[a]))return 0;
    }
  }
#if IVALUE==ILONG
  if(i>=4294967296){
#ifdef EXTENDED_PRIMES
    return miller_rabin(i,2) && 
      miller_rabin(i,325) &&
      miller_rabin(i,9375) &&
      miller_rabin(i,28178) &&
      miller_rabin(i,450775) &&
      miller_rabin(i,9780504) &&
      miller_rabin(i,1795265022);
#else
    fprintf(stderr,"miller-rabin: error, number n=%ld too big\n",i);
    return 0;
#endif
  }
#endif
  return miller_rabin(i,2) && miller_rabin(i,7) && miller_rabin(i,61);
}

int operator<(const value& v1, const value& v2){
  return v1.d<v2.d;
}

int operator>(const value& v1, const value& v2){
  return v1.d>v2.d;
}

int operator<=(const value& v1, const value& v2){
  return v1.d<=v2.d;
}

int operator>=(const value& v1, const value& v2){
  return v1.d>=v2.d;
}

int operator==(const value& v1, const value& v2){
  if(v1.type!=v2.type)return 0;
  if(v1.type==INTEGER)return v1.i==v2.i;
  else return v1.d==v2.d;
}

int operator!=(const value& v1, const value& v2){
  if(v1.type!=v2.type)return 1;
  if(v1.type==INTEGER)return v1.i!=v2.i;
  else return v1.d!=v2.d;
}

value value::lt(const value& v) const{
  return value((itype)(d<v.d));
}

value value::gt(const value& v) const{
  return value((itype)(d>v.d));
}

value value::le(const value& v) const{
  return value((itype)(d<=v.d));
}

value value::ge(const value& v) const{
  return value((itype)(d>=v.d));
}

value value::eq(const value& v) const{
  if(type!=v.type)return value((itype)0);
  if(type==INTEGER)return value((itype)(i==v.i));
  else return value((itype)(d==v.d));
}

value value::neq(const value& v) const{
  if(type!=v.type)return value((itype)1);
  if(type==INTEGER)return value((itype)(i!=v.i));
  else return value((itype)(d!=v.d));
}


void value::print(FILE* f, const char* Fd, const char* Ff){
  switch(type){
    case INTEGER:
      fprintf(f,Fd,i);
      return;
    case FLOATING:
      fprintf(f,Ff,d);
      return;
  }
}

const value VAL_PI=value(PI);
const value VAL_AA=value(AA);
const value VAL_DEG=value(DEG);
const value VAL_EMC=value(EMC);

const value VAL_ZERO=value((itype)0);
const value VAL_ONE=value((itype)1);
const value VAL_NEG=value((itype)(-1));

const value VAL_FZERO=value((ieee)0);

//declarations of operators that don't belong in the local_ops
//they are not used in string matching and must be available outside

//parentheses operators
opid shared::shared_op_parL=(opid){("{}"),OP_LIST     ,0,ASOC_MULTI};
opid shared::shared_op_par1=(opid){("[]"),OP_ITER1    ,0,ASOC_MULTI};
opid shared::shared_op_par2=(opid){("()"),OP_ITER2    ,0,ASOC_MULTI};
opid shared::shared_op_parF=(opid){("{::}"),OP_FUN    ,0,ASOC_MULTI};
//special case
opid shared::shared_op_minus=(opid){("-"),OP_MINUSSIGN ,11,ASOC_RUNARY};

//definition of local_ops
static opid local_ops[]={
  (opid){("!"),OP_FACT	,9,ASOC_LUNARY},	//0
  (opid){("#"),OP_HASH	,8,ASOC_RUNARY},	//1
  (opid){("$"),OP_TAKE	,9,ASOC_RUNARY},	//2
  (opid){("%"),OP_MOD	,5,ASOC_LINFIX},	//3
  (opid){("&"),OP_AND	,3,ASOC_LINFIX},	//4
  (opid){("*"),OP_MUL	,5,ASOC_LINFIX},	//5
  (opid){("+"),OP_PLUS	,4,ASOC_LINFIX},	//6
  (opid){("-"),OP_MINUS	,4,ASOC_LINFIX},	//7
  (opid){("/"),OP_DIV	,5,ASOC_LINFIX},	//8
  (opid){("<"),OP_LT	,1,ASOC_LINFIX},	//9
  (opid){("="),OP_EQUALS	,1,ASOC_LINFIX},	//10
  (opid){(">"),OP_GT	,1,ASOC_LINFIX},	//11
  (opid){("?"),OP_MASK	,8,ASOC_RUNARY},	//12
  (opid){("@"),OP_PUT	,9,ASOC_RBINARY},	//13
  (opid){("D"),OP_DOUBLE	,8,ASOC_RUNARY},	//14
  (opid){("E"),OP_EVEN	,8,ASOC_RUNARY},	//15
  (opid){("F"),OP_FACTOR	,8,ASOC_RUNARY},	//16
  (opid){("I"),OP_INTEGER	,8,ASOC_RUNARY},	//17
  (opid){("N"),OP_NEGATIVE	,8,ASOC_RUNARY},	//18
  (opid){("O"),OP_ODD	,8,ASOC_RUNARY},	//19
  (opid){("P"),OP_POSITIVE	,8,ASOC_RUNARY},	//20
  (opid){("Q"),OP_SYSCALL	,9,ASOC_RUNARY},	//21
  (opid){("S"),OP_SORT	,8,ASOC_RUNARY},	//22
  (opid){("U"),OP_UNIQUE	,8,ASOC_RUNARY},	//23
  (opid){("X"),OP_PRIME	,8,ASOC_RUNARY},	//24
  (opid){("\\"),OP_WITHOUT	,3,ASOC_LINFIX},	//25
  (opid){("^"),OP_POW	,6,ASOC_RINFIX},	//26
  (opid){("_"),OP_INDEX	,7,ASOC_LINFIX},	//27
  (opid){("c"),OP_NCK	,7,ASOC_RINFIX},	//28
  (opid){("d"),OP_FLOAT	,8,ASOC_RUNARY},	//29
  (opid){("e"),OP_FLE	,10,ASOC_RINFIX},	//30
  (opid){("f"),OP_FRAC	,8,ASOC_RUNARY},	//31
  (opid){("i"),OP_INT	,8,ASOC_RUNARY},	//32
  (opid){("r"),OP_ROUND	,8,ASOC_RUNARY},	//33
  (opid){("x"),OP_X	,11,ASOC_BLANK},	//34
  (opid){("|"),OP_OR	,2,ASOC_LINFIX},	//35
  (opid){("~"),OP_REVERSE	,8,ASOC_RUNARY},	//36

  (opid){("#_"),OP_COUNT	,8,ASOC_RBINARY},	//37
  (opid){("%%"),OP_XMOD	,5,ASOC_LINFIX},	//38
  (opid){("**"),OP_XMUL	,5,ASOC_LINFIX},	//39
  (opid){("++"),OP_XPLUS	,4,ASOC_LINFIX},	//40
  (opid){("--"),OP_XMINUS	,4,ASOC_LINFIX},	//41
  (opid){("//"),OP_XDIV	,5,ASOC_LINFIX},	//42
  (opid){(":%"),OP_MMOD	,5,ASOC_LINFIX},	//43
  (opid){(":*"),OP_MMUL	,5,ASOC_LINFIX},	//44
  (opid){(":+"),OP_PPLUS	,4,ASOC_LINFIX},	//45
  (opid){(":-"),OP_MMINUS	,4,ASOC_LINFIX},	//46
  (opid){(":/"),OP_DDIV	,5,ASOC_LINFIX},	//47
  (opid){("::"),OP_FCALL	,8,ASOC_RINFIX},	//48
  (opid){(":<"),OP_LLT	,1,ASOC_LINFIX},	//49
  (opid){(":>"),OP_GGT	,1,ASOC_LINFIX},	//50
  (opid){(":^"),OP_PPOW	,6,ASOC_RINFIX},	//51
  (opid){("<<"),OP_SHL	,7,ASOC_LINFIX},	//52
  (opid){("<="),OP_LEQ	,1,ASOC_LINFIX},	//53
  (opid){("<>"),OP_NEQ	,1,ASOC_LINFIX},	//54
  (opid){("=="),OP_EQ	,1,ASOC_LINFIX},	//55
  (opid){(">="),OP_GEQ	,1,ASOC_LINFIX},	//56
  (opid){(">>"),OP_SHR	,7,ASOC_LINFIX},	//57
  (opid){("@!"),OP_VNRM	,8,ASOC_RUNARY},	//58
  (opid){("@*"),OP_OTIMES	,8,ASOC_RUNARY},	//59
  (opid){("@+"),OP_OPLUS	,8,ASOC_RUNARY},	//60
  (opid){("@:"),OP_DOT	,5,ASOC_LINFIX},	//61
  (opid){("@="),OP_VLEN	,8,ASOC_RUNARY},	//62
  (opid){("@^"),OP_CROSS	,6,ASOC_RINFIX},	//63
  (opid){("aa"),OP_AA	,11,ASOC_BLANK},	//64
  (opid){("bj"),OP_BESJ	,8,ASOC_RBINARY},	//65
  (opid){("by"),OP_BESY	,8,ASOC_RBINARY},	//66
  (opid){("mp"),OP_MP	,11,ASOC_BLANK},	//67
  (opid){("pd"),OP_PD	,8,ASOC_RUNARY},	//68
  (opid){("pi"),OP_PI	,11,ASOC_BLANK},	//69
  (opid){("pm"),OP_PM	,11,ASOC_BLANK},	//70
  (opid){("px"),OP_PX	,8,ASOC_RBINARY},	//71
  (opid){("pz"),OP_PZ	,8,ASOC_RUNARY},	//72
  (opid){("|&"),OP_XOR	,3,ASOC_LINFIX},	//73

  (opid){("#__"),OP_HIST	,8,ASOC_RBINARY},	//74
  (opid){(":::"),OP_FITER	,8,ASOC_RINFIX},	//75
  (opid){(":<="),OP_LLEQ	,1,ASOC_LINFIX},	//76
  (opid){(":<>"),OP_NNEQ	,1,ASOC_LINFIX},	//77
  (opid){(":=="),OP_EEQ	,1,ASOC_LINFIX},	//78
  (opid){(":>="),OP_GGEQ	,1,ASOC_LINFIX},	//79
  (opid){("abs"),OP_ABS	,8,ASOC_RUNARY},	//80
  (opid){("acs"),OP_ACOS	,8,ASOC_RUNARY},	//81
  (opid){("asn"),OP_ASIN	,8,ASOC_RUNARY},	//82
  (opid){("atn"),OP_ATAN	,8,ASOC_RUNARY},	//83
  (opid){("avg"),OP_OMEAN	,8,ASOC_RUNARY},	//84
  (opid){("cbt"),OP_CBRT	,8,ASOC_RUNARY},	//85
  (opid){("cos"),OP_COS	,8,ASOC_RUNARY},	//86
  (opid){("deg"),OP_DEG	,11,ASOC_BLANK},	//87
  (opid){("emc"),OP_EMC	,11,ASOC_BLANK},	//88
  (opid){("exp"),OP_EXP	,8,ASOC_RUNARY},	//89
  (opid){("gcd"),OP_GCD	,8,ASOC_RUNARY},	//90
  (opid){("lcm"),OP_LCM	,8,ASOC_RUNARY},	//91
  (opid){("log"),OP_LOG	,8,ASOC_RUNARY},	//92
  (opid){("max"),OP_MAX	,8,ASOC_RUNARY},	//93
  (opid){("min"),OP_MIN	,8,ASOC_RUNARY},	//94
  (opid){("nrm"),OP_ONORM	,8,ASOC_RBINARY},	//95
  (opid){("phi"),OP_EFI	,8,ASOC_RUNARY},	//96
  (opid){("ran"),OP_RAND	,8,ASOC_RUNARY},	//97
  (opid){("sin"),OP_SIN	,8,ASOC_RUNARY},	//98
  (opid){("sqt"),OP_SQRT	,8,ASOC_RUNARY},	//99
  (opid){("tan"),OP_TAN	,8,ASOC_RUNARY},	//100
  (opid){("var"),OP_OVAR	,8,ASOC_RUNARY},	//101

  (opid){("acos"),OP_ACOS	,8,ASOC_RUNARY},	//102
  (opid){("asin"),OP_ASIN	,8,ASOC_RUNARY},	//103
  (opid){("atan"),OP_ATAN	,8,ASOC_RUNARY},	//104
  (opid){("cbrt"),OP_CBRT	,8,ASOC_RUNARY},	//105
  (opid){("sqrt"),OP_SQRT	,8,ASOC_RUNARY},	//106
};

static opid* local_ops_slices[OPS_LEN_MAX+1]={
  0,
  local_ops,
  local_ops+37,
  local_ops+74,
  local_ops+102
};

static uint local_ops_lens[OPS_LEN_MAX+1]={
  0,37,37,28,5
};


#ifndef NOSYSCALL
int shared::special_interrupt=INTERRUPT_NONE;
#endif
int shared::exit_returnvalue=0;

#ifdef SIGHANDLE
sig_atomic_t shared::interactive_flag=0;
sig_atomic_t shared::state=STATE_IDLE;

sig_atomic_t shared::interrupt=0; //volatile?
char* shared::wait_ptr=0;
#endif

//returns the length of the array
int findop(char* c, opid& op){
  for(int len=OPS_LEN_MAX;len>0;--len){
    uint min=0;
    uint max=local_ops_lens[len]-1;
    int res;
    res=strncmp(c,local_ops_slices[len][min].name,len);
    if(!res){
      op=local_ops_slices[len][min];
      return len;
    }
    if(res<0)continue;
    res=strncmp(c,local_ops_slices[len][max].name,len);
    if(!res){
      op=local_ops_slices[len][max];
      return len;
    }
    if(res>0)continue;
    while(max-min>1){
      uint mid=(min+max)/2;
      res=strncmp(c,local_ops_slices[len][mid].name,len);
      if(!res){
	op=local_ops_slices[len][mid];
	return len;
      }
      if(res>0){
        min=mid;
      }else{
        max=mid;
      }
    }
  }
  return 0;
}
