/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#ifndef H_BASE
#define H_BASE

#include "config.h"

#define DEBUG 0
#define SIGHANDLE
#define STOREHISTORY

#ifndef TRAPOVERFLOWS
#define TRAPOVERFLOWS 1
#endif

#if TOGGLESYSCALL==0
#define NOSYSCALL
#else
#ifdef NOSYSCALL
#undef NOSYSCALL
#endif
#endif

//#define TIMERAND

#include <cmath>
#include <cstdlib>
#include <cstdio>

#include <vector>

#ifdef SIGHANDLE
#include <csignal>
#endif

using namespace std;

#define FLOATING 1
#define INTEGER  2

//declare macros for different sizes of float and int values
#define FLOAT   4
#define DOUBLE  8
#define LONG   12

#define IINT    4
#define ILONG   8
//#define I128BIT 16

//in case it wasn't declared by configure script
#ifndef VALUE
#define VALUE DOUBLE
#endif

#ifndef IVALUE
#define IVALUE IINT
#endif



#if VALUE==FLOAT
typedef float ieee;
#define sqrt sqrtf
#define exp expf
#define log logf
#define sin sinf
#define cos cosf
#define tan tanf
#define sinh sinhf
#define cosh coshf
#define asin asinf
#define acos acosf
#define atan atanf
#define fabs fabsf
#define log10 log10f
#define ceil ceilf
#define floor floorf
#define pow powf
#define atan2 atan2f
#define cbrt cbrtf
#define jn jnf
#define yn ynf
#define lgamma lgammaf
#define lgamma_r lgammaf_r
#define erf erff
#define exp10 exp10f
#elif VALUE==DOUBLE
typedef double ieee;
#elif VALUE==LONG
typedef long double ieee;
#define sqrt sqrtl
#define exp expl
#define log logl
#define sin sinl
#define cos cosl
#define tan tanl
#define sinh sinhl
#define cosh coshl
#define asin asinl
#define acos acosl
#define atan atanl
#define fabs fabsl
#define log10 log10l
#define ceil ceill
#define floor floorl
#define pow powl
#define atan2 atan2l
#define cbrt cbrtl
#define jn jnl
#define yn ynl
#define lgamma lgammal
#define lgamma_r lgammal_r
#define erf erfl
#define exp10 exp10l
#endif

//in case exp10-like functions are not defined
#ifndef HAVE_EXP10_SET
#define LOG10 ((ieee)2.30258509299404568401799145468)
ieee exp10(ieee x);
#endif

#define PI   ((ieee)3.1415926535897932384626433832795)
#define AA   ((ieee)(-0.76666469596212309311120442251031))
#define DEG  ((ieee)0.0174532925199432957692369076849)
#define EMC  ((ieee)0.57721566490153286060651209008240243)

#ifndef NAN
#define NAN (0./0.)
#endif

typedef unsigned char uchar;
typedef unsigned int uint;


#if IVALUE==IINT
typedef int itype;
#elif IVALUE==ILONG
typedef long long itype;
//#elif IVALUE==I128BIT
//typedef __int128 itype;
#endif

#undef TRAP_OVR

#if TRAPOVERFLOWS
#ifdef __x86_64__

#if IVALUE==IINT

#ifdef __SIZEOF_INT__
#if __SIZEOF_INT__ == 4
#define TRAP_OVR
#endif // 4
#endif // __SIZEOF_INT__

#elif IVALUE==ILONG

#ifdef __SIZEOF_LONG_LONG__
#if __SIZEOF_LONG_LONG__ == 8
#define TRAP_OVR
#endif // 8
#endif // __SIZEOF_LONG_LONG__

#endif //IVALUE

#endif //__x86_64__
#endif //TRAPOVERFLOWS

#ifndef uint64_t
typedef unsigned long long uint64_t; //might not be right but close enough
#endif

//if we are on x86_64, we can do modular multiplication without overflow
#ifdef __x86_64__
#if IVALUE==ILONG
#if __SIZEOF_LONG_LONG__ == 8 //if this is not true, we have problems
#define EXTENDED_PRIMES
#endif
#endif
#endif

struct value {
  uint type;
  itype i;
  ieee d;
  itype assertI(const char* op);
  ieee  assertD(const char* op);
  value operator+(const value& v) const;
  value operator-(const value& v) const;
  value operator*(const value& v) const;
  value operator/(const value& v) const;
  value operator^(const value& v) const;
  value operator%(const value& v) const;
  value e(const value& v) const;
  value binomial(const value& v) const;

  void operator+=(const value& v);
  void operator-=(const value& v);
  void operator*=(const value& v);
  void operator/=(const value& v);


  value operator+(const ieee& v) const;
  value operator-(const ieee& v) const;
  value operator*(const ieee& v) const;
  value operator/(const ieee& v) const;
  value operator^(const ieee& v) const;
  value operator%(const ieee& v) const;
  value operator+(const itype& v) const;
  value operator-(const itype& v) const;
  value operator*(const itype& v) const;
  value operator/(const itype& v) const;
  value operator^(const itype& v) const;
  value operator%(const itype& v) const;
  value operator-() const;
  void op_neg();
  void op_sin();
  void op_cos();
  void op_tan();
  void op_atan();
  void op_asin();
  void op_acos();
  void op_abs();
  void op_exp();
  void op_sqrt();
  void op_log();
  void op_cbrt();
  void op_int();
  void op_float();
  void op_frac();
  void op_round();
  void op_fact();
  void op_exp10();
  void op_rand();
  value op_bj(int n);
  value op_by(int n);
  int is_odd() const;
  int is_even() const;
  int is_positive() const;
  int is_negative() const;
  int is_int() const;
  int is_float() const;
  int is_prime() const;

  value lt(const value& v) const;
  value gt(const value& v) const;
  value le(const value& v) const;
  value ge(const value& v) const;
  value eq(const value& v) const;
  value neq(const value& v) const;


  itype intpart() const;

  value(){
    type=0;
  }
  value(itype ii){
    type=INTEGER;
    i=ii;
    d=ii;
  }
  value(ieee ff){
    type=FLOATING;
    d=ff;
  }
  value(const char* s);
  void print(FILE* f, const char* Fd, const char* Ff);
};

extern const value VAL_PI;
extern const value VAL_AA;
extern const value VAL_DEG;
extern const value VAL_EMC;

extern const value VAL_ZERO;
extern const value VAL_ONE;
extern const value VAL_NEG;

extern const value VAL_FZERO;

//external, useful as comparators for sort, min, max
int operator<(const value& v1, const value& v2);
int operator>(const value& v1, const value& v2);
int operator<=(const value& v1, const value& v2);
int operator>=(const value& v1, const value& v2);
int operator==(const value& v1, const value& v2);
int operator!=(const value& v1, const value& v2);

//asociativnost
#define ASOC_BLANK   0x00
#define ASOC_RUNARY  0x01
#define ASOC_LUNARY  0x02
#define ASOC_RINFIX  0x03
#define ASOC_LINFIX  0x04
#define ASOC_RBINARY 0x05
#define ASOC_MULTI   0x10


#define OPMASK 0xFF00

#define OP_NONE 0
//memory
#define OPX_MEM    0x0000
#define OP_TAKE    0x0001
#define OP_PUT     0x0002
#define OP_SYSCALL 0x0003
#define OP_FCALL   0x0004
#define OP_FITER   0x0005
#define OP_X       0x0006

//component-wise
#define OPX_COM    0x0100
#define OP_PPLUS   0x0101
#define OP_MMINUS  0x0102
#define OP_MMUL    0x0103
#define OP_DDIV    0x0104
#define OP_MMOD    0x0105
#define OP_PPOW    0x0106

//accumulators
#define OPX_ACC    0x0200
#define OP_OPLUS   0x0201
#define OP_OTIMES  0x0202
#define OP_OMEAN   0x0203
#define OP_OVAR    0x0204
#define OP_ONORM   0x0205

//function maps
#define OPX_FUN      0x0300
#define OP_FACT      0x0301
#define OP_SIN       0x0302
#define OP_COS       0x0303
#define OP_TAN       0x0304
#define OP_ASIN      0x0305
#define OP_ACOS      0x0306
#define OP_ATAN      0x0307
#define OP_SQRT      0x0308
#define OP_CBRT      0x0309
#define OP_EXP       0x030A
#define OP_LOG       0x030B
#define OP_MINUSSIGN 0x030C
#define OP_ABS       0x030D
#define OP_ROUND     0x030E
#define OP_FRAC      0x030F
#define OP_INT       0x0310
#define OP_FLOAT     0x0311
#define OP_RAND      0x0312

//function maps with argument
#define OPX_FUNX     0x0400
#define OP_BESJ      0x0401
#define OP_BESY      0x0402

//tensor operators
#define OPX_TEN      0x0500
#define OP_PLUS      0x0501
#define OP_MINUS     0x0502
#define OP_MUL       0x0503
#define OP_DIV       0x0504
#define OP_MOD       0x0505
#define OP_POW       0x0506
#define OP_FLE       0x0507
#define OP_NCK       0x0508

//set-wise
#define OPX_SET      0x0600
#define OP_AND       0x0601
#define OP_WITHOUT   0x0602
#define OP_OR        0x0603
#define OP_XOR       0x0604
#define OP_REVERSE   0x0605
#define OP_SORT      0x0606
#define OP_HASH      0x0607
#define OP_MASK      0x0608
#define OP_EQUALS    0x0609
#define OP_COUNT     0x060A
#define OP_HIST      0x060B

//set-select
#define OPX_SEL      0x0700
#define OP_SHL       0x0701
#define OP_SHR       0x0702
#define OP_INDEX     0x0703
#define OP_UNIQUE    0x0704
#define OP_MIN       0x0705
#define OP_MAX       0x0706

//set-filter
#define OPX_FILT     0x0800
#define OP_POSITIVE  0x0801
#define OP_NEGATIVE  0x0802
#define OP_ODD       0x0803
#define OP_EVEN      0x0804
#define OP_INTEGER   0x0805
#define OP_DOUBLE    0x0806
#define OP_PRIME     0x0807

//component-wise logic
#define OPX_CLOG     0x0900
#define OP_LLT       0x0901
#define OP_GGT       0x0902
#define OP_LLEQ      0x0903
#define OP_GGEQ      0x0904
#define OP_EEQ       0x0905
#define OP_NNEQ      0x0906

//tensor logic
#define OPX_TLOG     0x0A00
#define OP_LT        0x0A01
#define OP_GT        0x0A02
#define OP_LEQ       0x0A03
#define OP_GEQ       0x0A04
#define OP_EQ        0x0A05
#define OP_NEQ       0x0A06

//const
#define OPX_CONST    0x0B00
#define OP_PI        0x0B01
#define OP_AA        0x0B02
#define OP_DEG       0x0B03
#define OP_EMC       0x0B04
#define OP_PM        0x0B05
#define OP_MP        0x0B06

//polynomial
#define OPX_POLY     0x0C00
#define OP_XPLUS     0x0C01
#define OP_XMINUS    0x0C02
#define OP_XMUL      0x0C03
#define OP_XDIV      0x0C04
#define OP_XMOD      0x0C05
#define OP_XPOW      0x0C06
#define OP_PX        0x0C07
#define OP_PD        0x0C08
#define OP_PZ        0x0C09

//vectors
#define OPX_VEC      0x0D00
#define OP_DOT       0x0D01
#define OP_VLEN      0x0D02
#define OP_VNRM      0x0D03
#define OP_CROSS     0x0D04

//number theory
#define OPX_NUMBR    0x0E00
#define OP_GCD	     0x0E01
#define OP_LCM	     0x0E02
#define OP_FACTOR    0x0E03
#define OP_EFI       0x0E04

//parentheses
#define OPX_PAREN    0xFF00
#define OP_LIST      0xFF0F
#define OP_ITER1     0xFF0E
#define OP_ITER2     0xFF0D
#define OP_FUN       0xFF0C


//syscall constants
#define SYSC_NOOP         0
#define SYSC_EXIT         1

#define SYSC_FLOW_BRK     100
#define SYSC_FLOW_CONT    101
#define SYSC_FLOW_PC      102
#define SYSC_FLOW_PCSET   103
#define SYSC_FLOW_STCKLVL 104
#define SYSC_FLOW_MAXIT   105

#define SYSC_HIST_COUNT   1000
#define SYSC_HIST_ENABLE  1001
#define SYSC_HIST_CLEAR   1002

#define SYSC_MEM_LCPOS    2000
#define SYSC_MEM_LPOS     2001
#define SYSC_MEM_AOFF     2002
#define SYSC_MEM_CLEAR    2003
#define SYSC_MEM_LIST     2004

#define SYSC_FUN_CLEAR    2050
#define SYSC_FUN_LIST     2051

#define SYSC_BLD_VERSION  3000
#define SYSC_BLD_SINT     3001
#define SYSC_BLD_SFLT     3002
#define SYSC_BLD_DEBUG    3003
#define SYSC_BLD_RL       3004
#define SYSC_BLD_OVERFLOW 3005

#define SYSC_SYS_TIME     5000
#define SYSC_SYS_PTIME    5001
#define SYSC_SYS_SLEEP    5002
#define SYSC_SYS_PID      5050

#define SYSCTIME_EPOCH 0
#define SYSCTIME_SEC   1
#define SYSCTIME_MIN   2
#define SYSCTIME_HOUR  3
#define SYSCTIME_MDAY  4
#define SYSCTIME_MON   5
#define SYSCTIME_YEAR  6
#define SYSCTIME_WDAY  7
#define SYSCTIME_YDAY  8

struct opid {
  const char* name;
  uint code;
  uchar precedence;
  uchar asoc;
};


#ifndef NOSYSCALL
#define INTERRUPT_NONE  0
#define INTERRUPT_EXIT  1
#define INTERRUPT_BREAK 2
#define INTERRUPT_CONT  3
#endif

#define STATE_IDLE 0
#define STATE_WORK 1

#define OPS_LEN_MAX 4

namespace shared {
#ifndef NOSYSCALL
  extern int special_interrupt;
#endif
  extern int exit_returnvalue;
  extern opid shared_op_parL;
  extern opid shared_op_par1;
  extern opid shared_op_par2;
  extern opid shared_op_parF;
  extern opid shared_op_minus;
#ifdef SIGHANDLE
  extern sig_atomic_t interactive_flag;
  extern sig_atomic_t state;
  extern sig_atomic_t interrupt;
  extern char* wait_ptr;
#endif
}

int findop(char* c, opid& op);

//assembler routines for trapping overflows
#ifdef TRAP_OVR
inline static void iadd_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"addl %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"addq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}

inline static void isub_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"subl %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"subq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}

inline static void imul_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"imull %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"imulq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}
#endif

#ifdef EXTENDED_PRIMES
inline static uint64_t imulmod_(uint64_t a, uint64_t b, uint64_t m){
  // Perform 128 multiplication and division
  uint64_t q; // q = floor(a b / m)
  uint64_t r; // r = a b mod m
  __asm__ __volatile__ (
			"mulq %3\n\t"
			"divq %4\n"
			:"=a"(q), "=d"(r)
			: "a"(a), "rm"(b), "rm"(m)
			);
  return r;
}

#endif
//end of assembler routines


#endif
