/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "methods.h"

using namespace std;

struct poly {
  int size;
  ieee* data;
  void alloc(int n){
    data=(ieee*)malloc(n*sizeof(ieee));
  }
  void dealloc(){
    free(data);
  }
  ieee& operator[](int i){
    return data[i];
  }
};

static void bairstow_getp(poly& p, ieee u, ieee v, ieee& c, ieee& d, ieee& g, ieee& h){
  int n=p.size-1;
  ieee b1=0;
  ieee b2=0;
  ieee f1=0;
  ieee f2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=p[i+2]-u*b1-v*b2;
    ieee tmpf=b2-u*f1-v*f2;
    b2=b1;
    b1=tmpb;
    f2=f1;
    f1=tmpf;
  }
  c=p[1]-u*b1-v*b2;
  d=p[0]-v*b1;
  g=b2-u*f1-v*f2;
  h=b1-v*f1;
}

static int bairstow_iter(poly& p, ieee& u, ieee& v, ieee& error,int limit){
  //newton iteration
  ieee c,d,g,h;
  ieee det;
  ieee err2;

  int count=0;
  while(count<limit){
    bairstow_getp(p,u,v,c,d,g,h);
    det=v*g*g+h*h-h*u*g;
    if(fabs(det)<1e-16){
      if((fabs(c)<1e-15)&&(fabs(d)<1e-15))break;//we hit the roots
      u+=AA/10;
      v+=PI/10;
      #if DEBUG>0
      fprintf(stderr,"bairstow solver: zero determinant occurred, retrying with u=%g, v=%g\n",(double)u,(double)v);
      #endif
      count++;
      continue;
    }
    det=1/det;
    ieee du=det*(-h*c+g*d);
    ieee dv=det*(-g*v*c+(g*u-h)*d);
    u-=du;
    v-=dv;
    error=sqrt(du*du+dv*dv);
    err2=sqrt(c*c+d*d);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: iteration step %d; error=%g, error(cd)=%g\n",count,(double)error,(double)err2);
    #endif
    if((error<1e-15)&&(err2<1e-9))break;
    count++;
  }
  if(err2>1e-9)error=1.;
  return count;
}

//"p" and "lower" must be allocated to accept polynomial extended to the next even order
static int bairstow_step(poly& p, poly& lower, array& out, int& addedzero){
  int n=p.size-1;
  #if DEBUG>0
  fprintf(stderr,"bairstow solver: starting step; n=%d\n",n);
  #endif
  //initial approximation
  ieee u=0;
  ieee v=-pow(fabs(p[0]/p[n]),2./n);

  #if DEBUG>0
  fprintf(stderr,"bairstow solver: initial approximation; u=%g, v=%g\n",(double)u,(double)v);
  #endif
  ieee error;
  int limit=n*24;
  int count=bairstow_iter(p, u, v, error,limit);

  if((count==limit)&&(error>1e-9)){
    u=p[n-1]/p[n];
    v=p[n-2]/p[n];
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: failed once, trying different approach; u=%g, v=%g\n",(double)u,(double)v);
    #endif
    count=bairstow_iter(p, u, v, error,limit);
  }

  if((count==limit)&&(error>1e-9)&&(n%2==1)){
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: failed twice, making polynomial of even order\n");
    #endif
    if(out.size()>0){
      ieee val=out.back().d;
      out.pop_back();
      //multiply poly by {-val 1}
      p[n+1]=p[n];
      p.size++;
      for(int i=n;i>0;i--){
	p[i]=p[i-1]-p[i]*val;
      }
      p[0]*=-val;
    }else{
      p[n+1]=p[n];
      p.size++;
      for(int i=n;i>0;i--){
	p[i]=p[i-1];
      }
      p[0]=0;
      addedzero=1;
#if DEBUG>0
      fprintf(stderr,"bairstow solver: added artificial zero\n");
#endif
    }
    n++;
    u=0;
    v=-pow(fabs(p[0]/p[n]),2./n);
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: new approximation; u=%g, v=%g\n",(double)u,(double)v);
    #endif
    count=bairstow_iter(p, u, v, error,limit);
  }

  if(count==limit){
    if(error<1e-13){
    }else
    if(error<1e-11){
      fprintf(stderr,"bairstow solver: low accuracy (%g), possibly multiple root\n",(double)error);
    }else if(error<1e-7){
      fprintf(stderr,"bairstow solver: seriously low accuracy (%g), possibly multiple root\n",(double)error);
    }else{
      fprintf(stderr,"bairstow solver: failed to converge in %d steps\n",limit);
      return 1;
    }
  }
  //divide to obtain lower-order polynomial
  lower.size=n-1;
  ieee b1=0;
  ieee b2=0;
  for(int i=n-2;i>=0;i--){
    ieee tmpb=p[i+2]-u*b1-v*b2;
    lower[i]=tmpb;
    b2=b1;
    b1=tmpb;
  }
    #if DEBUG>0
    fprintf(stderr,"bairstow solver: successfully divided\n");
    #endif
  //found roots
  ieee disc=u*u-4*v;
  if(disc<-1e-14*u*u)return 0;
  if(disc<0)disc=0;
  out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
  out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
  return 0;
}

int bairstow(array& polynomial, array& out){
  int n2=polynomial.size();
  int n1=0;
  //trim zeroes
  for(int i=n2-1;i>=0;i--){
    if(polynomial[i].d==0){
      n2=i;
    }else break;
  }
  //front zeroes
  for(int i=0;i<n2;i++){
    if(polynomial[i].d==0){
      out.push_back(value((ieee)0));
      n1=i+1;
    }else{
      break;
    }
  }
  if(n2-n1>12){
    fprintf(stderr,"bairstow solver: warning, polynomial order is high (%d)\n",n2-n1-1);
  }
  switch(n2-n1){
    case 0:return 0;//all reals are solutions
    case 1:{//no zeroes
      return 0;
    }
    case 2:{
      out.push_back(value(-polynomial[n1].d/polynomial[n1+1].d));
      return 0;
    }
    case 3:{
      ieee u=polynomial[n1+1].d/polynomial[n1+2].d;
      ieee v=polynomial[n1].d/polynomial[n1+2].d;
      ieee disc=u*u-4*v;
      if(disc<-1e-14*u*u)return 0;
      if(disc<0)disc=0;
      out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
      out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
      return 0;
    }
    default:{
      int addedzero=0;
      poly p;
      p.alloc(n2-n1+1);
      for(int i=n1;i<n2;i++){
	p[i-n1]=polynomial[i].d/polynomial[n2-1].d;
      }
      p.size=n2-n1;
      poly p2;
      p2.alloc(n2-n1+1);
      while(p.size>3){
	if(bairstow_step(p,p2,out,addedzero)){
	  if(addedzero)out.clear();
	  p.dealloc();
	  p2.dealloc();
	  return 1;
	}else{
	  poly p3=p;
	  p=p2;
	  p2=p3;
	}
      }
      //now we got polynomial of order 1 or 2:
      if(p.size==2){
	out.push_back(value(-p[0]/p[1]));
      }else{
	ieee u=p[1]/p[2];
	ieee v=p[0]/p[2];
	ieee disc=u*u-4*v;
	if(disc>-1e-14*u*u){
	  if(disc<0)disc=0;
	  out.push_back(value(((ieee)0.5)*(-u+sqrt(disc))));
	  out.push_back(value(((ieee)0.5)*(-u-sqrt(disc))));
	}
      }
      p.dealloc();
      p2.dealloc();
      if(addedzero){
#if DEBUG>0
	fprintf(stderr,"bairstow solver: cleaning up added zero\n");
#endif
	int minindex=-1;
	ieee minz=1e-3;
	for(int j=0;j<out.size();j++){
	  ieee fbs=fabs(out[j].d);
	  if(fbs<minz){
	    minindex=j;
	    minz=fbs;
	  }
	}
	if(minindex>-1){
	  out.erase(minindex);
	}
	else{
	  fprintf(stderr,"bairstow solver: warning, added zero disappeared, one root may be wrong or missing\n");
	}
      }
      return 0;
    }
  }
}

void trimzeroes(array& a){
  while(a.size()&&(a.back().d==0)){
    a.pop_back();
  }
}

int polydivmod(array& a, array& b, array& divmod){
  int na=a.size();
  int nb=b.size();
  int n=na-nb+1;
  divmod.transfer(a);
  if(na<nb){//not divisible, everything is remainder
    return na;
  }
  for(int i=n-1;i>=0;i--){
    if(divmod[i+nb-1].d==0)continue;
    divmod[i+nb-1]=divmod[i+nb-1]/b[nb-1].d;
    for(int j=0;j<nb-1;j++){
      divmod[i+j]=divmod[i+j]-b[j]*divmod[i+nb-1].d;
    }
  }
  return na-n;
}


//code from wikipedia
//takes already positive values
static itype gcd_pair(itype u, itype v){
  int shift;
  /* GCD(0,v) == v; GCD(u,0) == u, GCD(0,0) == 0 */
  if (!u)return v;
  if (!v)return u;
 
  /* Let shift := lg K, where K is the greatest power of 2
        dividing both u and v. */
  for (shift = 0; ((u | v) & 1) == 0; ++shift) {
    u >>= 1;
    v >>= 1;
  }
 
  while ((u & 1) == 0)u >>= 1;
 
  // From here on, u is always odd.
  do{
    // remove all factors of 2 in v -- they are not common
    //   note: v is not zero, so while will terminate
    while((v & 1) == 0)v >>= 1;
    if (u > v) {
      itype t = v; v = u; u = t;
    }  // Swap u and v.
    v-=u;
  } while (v);
  // restore common factors of 2
  return u << shift;
}

//one could precompute common shift
void gcd(array& a, array& out){
  if(!a.size())return;
  int presign=1;
  //test if all the operands are integers and factor out the signs (even number of minuses yields positive result)
  for(array::iterator it=a.begin();it.ready();++it){
    if((*it).type!=INTEGER){
      fprintf(stderr,"gcd: warning, noninteger value encountered, returning empty array\n");
      return;
    }
    if((*it).i<0){
      presign=-presign;
      (*it).i=-(*it).i;
      (*it).d=-(*it).d;//not needed but here for consistency
    }
  }
  if(a.size()==1){
    out.push_back(a[0]);
    return;
  }
  //perform the algorithm
  itype result=a[0].i;
  for(array::iterator it=a.get(1);it.ready();++it){
    result=gcd_pair(result,(*it).i);
  }
  out.push_back(presign*result);
}

void lcm(array& a, array& out){
  int nonzero=1;
  if(!a.size())return;
  //test if all the operands are integers and factor out the signs (even number of minuses yields positive result)
  for(array::iterator it=a.begin();it.ready();++it){
    if((*it).type!=INTEGER){
      fprintf(stderr,"lcm: warning, noninteger value encountered, returning empty array\n");
      return;
    }
    if((*it).i<0){
      (*it).i=-(*it).i;
      (*it).d=-(*it).d;//not needed but here for consistency
    }
    if((*it).i==0)nonzero=0;
  }
  if(a.size()==1){
    out.push_back(a[0]);
    return;
  }
  if(!nonzero){
    out.push_back(VAL_ZERO);
    return;
  }
  //perform the algorithm
  itype result=a[0].i;
  for(array::iterator it=a.get(1);it.ready();++it){
    itype b=(*it).i;
    itype g=gcd_pair(result,b);
#ifdef TRAP_OVR
    char flag=0;
    imul_(result,b/g,flag);
    if(flag){
      fprintf(stderr,"runtime: integer overflow on multiplication in lcm{...}, returning empty array\n");
      return;
    }
#else
    result*=(b/g);
#endif
  }
  out.push_back(result);
}


void pfactor(itype p, array& out){
  static const uint jumptable[]={6,4,2,4,2,4,6,2};
  if(p==0){
    out.push_back(VAL_ZERO);
    return;
  }else if(p<0){
    out.push_back(VAL_NEG);
    p=-p;
  }
  if(p==1)return;//not valid prime
  while(!(p&1)){//remove twos
    p>>=1;
    out.push_back((itype)2);
  }
  while(!(p%3)){//threes
    p/=3;
    out.push_back((itype)3);
  }
  while(!(p%5)){//fives
    p/=5;
    out.push_back((itype)5);
  }
  if(value(p).is_prime()){//if it is prime, don't bother
    out.push_back(p);
    return;
  }
  uint index=1;
  for(uint64_t i=7;i*i<=p;i+=jumptable[0x7&(index++)]){
    int flag=0;

    //alternative to modulus
    /*
    while(1){
      uint64_t mod=p;
      uint64_t quot=0;
      uint64_t divby=i;
      uint64_t mask=1;
      while(mod>=divby){
	if(mod&mask){
	  quot|=mask;
	  mod-=divby;
	}
	divby<<=1;
	mask<<=1;
      }
      if(!mod){
	p=quot;
	out.push_back((itype)i);
	flag=1;
      }else{
	break;
      }
    }
    */

    while(!(p%i)){
      p/=i;
      out.push_back((itype)i);
      flag=1;
    }

    if(flag && value(p).is_prime())break;
#ifdef SIGHANDLE
    if(shared::interrupt){break;}
#endif
  }
  if(p>1){
    out.push_back(p);
  }
}

itype onephi(itype n){
  static const uint jumptable[]={6,4,2,4,2,4,6,2};
  if(n<3)return 1;
  uint result=1;
  uint more_than_once=0;
  while(!(n&1)){//remove twos
    n>>=1;
    if(more_than_once){
      result*=2;
    }else{
      more_than_once=1;
    }
  }
  more_than_once=0;
  while(!(n%3)){//threes
    n/=3;
    if(more_than_once){
      result*=3;
    }else{
      result*=2;
      more_than_once=1;
    }
  }
  more_than_once=0;
  while(!(n%5)){//fives
    n/=5;
    if(more_than_once){
      result*=5;
    }else{
      result*=4 ;
      more_than_once=1;
    }
  }
  if(value(n).is_prime()){//if it is prime, don't bother
    result*=n-1;
    return result;
  }
  uint index=1;
  for(uint i=7;i*i<n;i+=jumptable[0x7&(index++)]){
    int flag=0;
    more_than_once=0;
    while(!(n%i)){
    flag=1;
      n/=i;
      if(more_than_once){
	result*=i;
      }else{
	result*=i-1;
	more_than_once=1;
      }
    }
    if(flag && value(n).is_prime())break;
#ifdef SIGHANDLE
    if(shared::interrupt){break;}
#endif
  }
  if(n>1){
    result*=n-1;
  }
  return result;
}
