/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <algorithm>
#include "base.h"
#include "unroll.h"

using namespace std;

//#define CONSTRUCTOR_PRINT

unroll::chunk* unroll::pool=0;

unroll::chunk* unroll::ch_alloc(size_t n){
  unroll::chunk* ch;
  if(n==CHUNKSIZE && pool){
#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"recycled (%d)\n",n);
#endif

    ch=pool;
    pool=pool->next;
    ch->count=0;
  }else{

#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"allocated (%d)\n",n);
#endif

    ch=(chunk*)malloc(sizeof(chunk)+n*sizeof(mytype));
    ch->size=n;
    ch->count=0;
  }
  return ch;
}

void unroll::ch_free(chunk* ch){
  if(ch->size==CHUNKSIZE){
#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"recycling (%d)\n",ch->size);
#endif
    ch->next=pool;
    pool=ch;
  }else{
#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"freeing (%d)\n",ch->size);
#endif

    free(ch);
  }
}

//add a desired size
void unroll::allocate(size_t n){
  if(!n)return;
  chunk* chnk=ch_alloc(n);
  if(chunkcount){
    chunk_last->next=chnk;
    chnk->prev=chunk_last;
    chunk_last=chnk;
  }else{
    chunk_first=chunk_last=chnk;
    chnk->prev=0;
  }
  chnk->next=0;
  ++chunkcount;
}

void unroll::expand(){
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"expanding\n");
#endif

  uint newsize=list_size/2;
  if(newsize<CHUNKSIZE)newsize=CHUNKSIZE;
  allocate(newsize);
}

//empty constructor doesn't do anything
unroll::unroll(){
  list_size=0;
  chunkcount=0;
  chunk_first=chunk_last=0;
  first.chnk=0;
  last.chnk=0;
}

//allocate
unroll::unroll(size_t newsize){
  list_size=0;
  chunkcount=0;
  first.chnk=0;
  last.chnk=0;
  if(!newsize){
    chunk_first=chunk_last=0;
  }else{
    allocate(newsize);
  }
}

void unroll::push_back(const mytype& val){
  if(!list_size){//first insertion; initialize iterators
    if(!chunkcount){//no chunks yet
      allocate(CHUNKSIZE);
    }
    first.initialize(chunk_first);
    last=first;
  }else{
    //last.chnk is nonnull because list is not empty
    if(last.pos<last.chnk->size-1){
      ++last.pos;
    }else{
      last.pos=0;
      if(!last.chnk->next)expand();
      last.chnk=last.chnk->next;
    }
  }
  *last=val;
  last.chnk->count++;
  list_size++;
}

void unroll::pop_back(){
  if(list_size){
    last.chnk->count--;
    --last;
    --list_size;
    if(!chunk_last->count){
      chunk* tofree=chunk_last;
      chunk_last=chunk_last->prev;
      if(chunk_last)chunk_last->next=0;
      ch_free(tofree);
      chunkcount--;
    }
    if(!list_size){
      first.chnk=0;
      last.chnk=0;
      if(!chunk_last)chunk_first=0;
    }
  }
}

mytype& unroll::back(){
  if(!list_size)fprintf(stderr,"runtime: empty list has no back\n");
  return *last;
}

mytype& unroll::front(){
  if(!list_size)fprintf(stderr,"runtime: empty list has no front\n");
  return *first;
}

void unroll::erase(uint i){
  if(i>=list_size){
    fprintf(stderr,"runtime: index out of bounds %u/%u\n",i,(uint)list_size);
    return;
  }
  chunk* ch=chunk_first;
  int cum=0;
  while(i>=cum+ch->count){
    cum+=ch->count;
    ch=ch->next;
  }
  //this is not tested yet
  int offset=i-cum;
  memmove(ch->data+offset,ch->data+(offset+1),sizeof(mytype)*(ch->count-offset-1));
  ch->count--;
  list_size--;
  --last;
  if(!list_size)first.chnk=last.chnk=0;
}

unroll::iterator unroll::get(uint i){
  if(i>=list_size){
    iterator ui;
    ui.chnk=0;
    ui.pos=0;
    return ui;
  }
  chunk* ch=chunk_first;
  int cum=0;
  while(i>=cum+ch->count){
    cum+=ch->count;
    ch=ch->next;
  }
  iterator ui;
  ui.chnk=ch;
  ui.pos=i-cum;
  return ui;
}

void unroll::append(unroll& ul){
  if(ul.list_size){//if not empty, append
    if(list_size && ul.chunkcount==1 && ul.list_size<=(last.chnk->size-last.chnk->count)){
      memcpy(last.chnk->data+last.chnk->count,ul.chunk_first->data,ul.list_size*sizeof(mytype));
      list_size+=ul.list_size;
      last.pos+=ul.list_size;
      last.chnk->count+=ul.list_size;
#ifdef CONSTRUCTOR_PRINT
      fprintf(stderr,"copy append\n");
#endif

    }else{

      if(chunkcount){
	//merge chunks
	chunk_last->next=ul.chunk_first;
	ul.chunk_first->prev=chunk_last;
      }else{
	//transfer chunks
	chunk_first=ul.chunk_first;
      }
      chunk_last=ul.chunk_last;
      chunkcount+=ul.chunkcount;

      //move front iterator
      if(!list_size){
	first=ul.first;
      }
      //move back iterator
      last=ul.last;
      list_size+=ul.list_size;


      //cleanup
      ul.chunkcount=0;
      ul.chunk_first=ul.chunk_last=0;
      ul.list_size=0;
      ul.first.chnk=0;
      ul.last.chnk=0;

#ifdef CONSTRUCTOR_PRINT
      fprintf(stderr,"transfer append\n");
#endif
      }
  }
}

void unroll::transfer(unroll& ul){
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"transfer\n");
#endif
  //free
  while(chunk_first){
    chunk* ch=chunk_first;
    chunk_first=chunk_first->next;
    ch_free(ch);
  }

  //transfer
  chunkcount=ul.chunkcount;
  list_size=ul.list_size;
  chunk_first=ul.chunk_first;
  chunk_last=ul.chunk_last;
  first=ul.first;
  last=ul.last;

  //mark as inexistent
  ul.chunkcount=0;
  ul.chunk_first=ul.chunk_last=0;
  ul.list_size=0;
  ul.first.chnk=0;
  ul.last.chnk=0;
}

void unroll::head(size_t count){
  if(count>=list_size)return;
  if(!count){clear();return;}

  last=get(count-1);//found the new last
  last.chnk->count=last.pos+1;//only so many elements
  list_size=count;//new number of elements

  //delete the rest
  chunk* todelete=last.chnk->next;
  last.chnk->next=0;
  while(todelete){
    chunk* after=todelete->next;
    ch_free(todelete);
    chunkcount--;
    todelete=after;
  }

}

void unroll::tail(size_t count){
  if(count>=list_size)return;
  if(!count){clear();return;}
  size_t delcount=list_size-count;

  int cum=0;
  while(delcount>=cum+chunk_first->count){//jump over; this deletes the chunk
    cum+=chunk_first->count;

    chunk* todelete=chunk_first;
    chunk_first=chunk_first->next;
    ch_free(todelete);
    chunkcount--;
  }
  chunk_first->prev=0;
  first.initialize(chunk_first);//initialization
  //move the memory (delcount-cum) slots to the left
  memmove(chunk_first->data,chunk_first->data+(delcount-cum),(chunk_first->count-(delcount-cum))*sizeof(mytype));
  chunk_first->count-=(delcount-cum);

  list_size=count;
  if(chunk_first==chunk_last){//restore
    last.pos=last.chnk->count-1;
  }
}


unroll::unroll(const unroll& ul){
  if(!ul.list_size){
#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"copy constructor (empty)\n");
#endif
    list_size=0;
    chunkcount=0;
    chunk_first=chunk_last=0;
    first.chnk=0;
    last.chnk=0;
    return;
  }
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"copy constructor\n");
#endif

  list_size=ul.list_size;
  chunkcount=0;
  allocate(list_size);
  first.chnk=chunk_first;
  first.pos=0;
  last.chnk=chunk_first;
  last.pos=list_size-1;
  //iterators are ready, now we copy data

  int cum=0;
  chunk* ch=ul.chunk_first;
  while(ch){
    memcpy(chunk_first->data+cum,ch->data,ch->count*sizeof(mytype));
    cum+=ch->count;
    ch=ch->next;
  }
  //cum==list_size now
  chunk_first->count=list_size;
}

void unroll::operator=(const unroll& ul){
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"copy = operator\n");
#endif
  //free
  while(chunk_first){
    chunk* ch=chunk_first;
    chunk_first=chunk_first->next;
    ch_free(ch);
  }

  if(!ul.list_size){
    list_size=0;
    chunkcount=0;
    chunk_first=chunk_last=0;
    first.chnk=0;
    last.chnk=0;
    return;
  }
  list_size=ul.list_size;
  chunkcount=0;
  allocate(list_size);
  first.chnk=chunk_first;
  first.pos=0;
  last.chnk=chunk_first;
  last.pos=list_size-1;
  //iterators are ready, now we copy data

  int cum=0;
  chunk* ch=ul.chunk_first;
  while(ch){
    memcpy(chunk_first->data+cum,ch->data,ch->count*sizeof(mytype));
    cum+=ch->count;
    ch=ch->next;
  }
  //cum==list_size now
  chunk_first->count=list_size;
}


mytype& unroll::operator[](int i){
  if(!list_size){
    fprintf(stderr,"runtime: attempted subscripting of empty list\n");
    return chunk_first->data[0];
  }
  if(i>=(int)list_size){
    i%=list_size;
  }else if(i<0){
    i=((int)list_size-(-i)%(int)list_size)%(int)list_size;
  }
  chunk* ch=chunk_first;
  int cum=0;
  while(i>=cum+ch->count){
    cum+=ch->count;
    ch=ch->next;
  }
  return ch->data[i-cum];
}

void unroll::reverse(){
  int count=list_size/2;
  iterator i1=first;
  iterator i2=last;
  for(int i=0;i<count;i++){
    mytype mt=*i1;
    *i1=*i2;
    *i2=mt;
    ++i1;
    --i2;
  }
}

void unroll::clear(){
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"clear (%d)\n",list_size);
#endif
  if(!list_size)return;
  chunk* ch=chunk_first;
  while(ch){
    ch->count=0;
    ch=ch->next;
  }
  first.chnk=0;
  last=first;
  list_size=0;
}

void unroll::sort(int (*compare)(const mytype&, const mytype&)){
  if(list_size<2){
    return;
  }
  unroll temporary(*this);

  mytype* iterator_first=temporary.chunk_first->data;
  mytype* iterator_last=iterator_first+list_size;

  stable_sort(iterator_first,iterator_last,compare);

  transfer(temporary);
}

mytype unroll::min(int (*compare)(const mytype&, const mytype&)){
  if(!list_size)return mytype();
  if(list_size==1){
    return *first;
  }
  mytype mymin;
  iterator ui=first;
  ++ui;
  if(compare(*first,*ui)){
    mymin=*first;
  }else{
    mymin=*ui;
  }
  for(++ui;ui.ready();++ui){
    if(!compare(mymin,*ui)){
      mymin=*ui;
    }
  }
  return mymin;
}

mytype unroll::accum(const mytype& start, void (mytype::*acc)(const mytype&)){
  if(!list_size)return start;
  mytype val=start;

  chunk* ch=chunk_first;
  while(ch){
    int cnt=ch->count;
    for(int i=0;i<cnt;i++){
      (val.*acc)(ch->data[i]);
    }
    ch=ch->next;
  }
  return val;
}

void unroll::map(void (mytype::*fun)()){
  if(!list_size)return;

  chunk* ch=chunk_first;
  while(ch){
    int cnt=ch->count;
    for(int i=0;i<cnt;i++){
      (ch->data[i].*fun)();
    }
    ch=ch->next;
  }
}

void unroll::components(unroll& ul1, unroll& ul2, mytype (mytype::*op)(const mytype&) const){
  size_t n1=ul1.list_size;
  size_t n2=ul2.list_size;
  size_t n=n1<n2?n1:n2;
  if(!n)return;
  allocate(n);

  array::iterator it1=ul1.begin();
  array::iterator it2=ul2.begin();
  for(int i=0;i<n;i++){
    push_back(((*it1).*op)(*it2));
    ++it1;
    ++it2;
  }
}

void unroll::tensor(unroll& ul1, unroll& ul2, mytype (mytype::*op)(const mytype&) const){
  size_t n1=ul1.list_size;
  size_t n2=ul2.list_size;
  size_t n=n1*n2;
  if(!n)return;
  allocate(n);

  chunk* ch=ul1.chunk_first;
  while(ch){
    int cnt=ch->count;
    for(int i=0;i<cnt;i++){
      chunk* ch2=ul2.chunk_first;
      while(ch2){
	int cnt2=ch2->count;
	for(int j=0;j<cnt2;j++){
	  push_back((ch->data[i].*op)(ch2->data[j]));
	}
	ch2=ch2->next;
      }
    }
    ch=ch->next;
  }
}


void unroll::filter(int (mytype::*fun)() const, unroll& out){
  if(!list_size)return;
  chunk* ch=chunk_first;
  while(ch){
    int cnt=ch->count;
    for(int i=0;i<cnt;i++){
      if((ch->data[i].*fun)())out.push_back(ch->data[i]);
    }
    ch=ch->next;
  }
}

int unroll::includes(const mytype& v){
  if(!list_size)return 0;
  chunk* ch=chunk_first;
  while(ch){
    int cnt=ch->count;
    for(int i=0;i<cnt;i++){
      if(ch->data[i]==v)return 1;
    }
    ch=ch->next;
  }
  return 0;
}

int unroll::operator==(const unroll& ul2){
  if(list_size!=ul2.list_size)return 0;
  array::iterator it1=first;
  array::iterator it2=ul2.first;
  for(int i=0;i<list_size;i++){
    if((*it1)!=(*it2)){
      return 0;
    }
    ++it1;
    ++it2;
  }
  return 1;
}


void unroll::straighten(){
  if(chunkcount>5){
#ifdef CONSTRUCTOR_PRINT
    fprintf(stderr,"straighten from %d\n",chunkcount);
#endif

    unroll ul(*this);
    transfer(ul);
  }
}

unroll::~unroll(){
#ifdef CONSTRUCTOR_PRINT
  fprintf(stderr,"destroy %d/%d\n",chunkcount,list_size);
#endif
  while(chunk_first){
    chunk* ch=chunk_first;
    chunk_first=chunk_first->next;
    ch_free(ch);
  }
}

#if DEBUG>1
void unroll::print(FILE* fd, const char* Fd, const char* Ff){
  fprintf(fd,"list #%u (%u)\n",(uint)chunkcount,(uint)list_size);
  chunk* cf=chunk_first;
  for(uint i=0;i<chunkcount;i++){
    fprintf(fd,"[%u] <%p|%p|%p> (%u/%u)",i,cf->prev,cf,cf->next,(uint)cf->count,(uint)cf->size);
    for(uint a=0;a<cf->count;a++){
      fputc(' ',fd);
      cf->data[a].print(fd,Fd,Ff);
    }
    fprintf(fd,"\n");
    cf=cf->next;
  }
}

void unroll::print_short(FILE* fd){
  fprintf(fd,"list #%u (%u)\n",(uint)chunkcount,(uint)list_size);
}

#endif
