/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#ifndef H_UNROLL
#define H_UNROLL

#define CHUNKSIZE 16

#include "base.h"

typedef value mytype;

class unroll {
 private:
  struct chunk {
    chunk* prev;
    chunk* next;
    size_t size;
    size_t count;
    mytype data[0];
  };

 private:

  static chunk* pool;
  static chunk* ch_alloc(size_t n);
  static void ch_free(chunk* ch);


 public:
  struct iterator {
    chunk* chnk;
    size_t pos;
    void initialize(chunk* ch){
      pos=0;
      chnk=ch;
    }
    void operator--(){//decrement
      if(pos==0){
	chnk=chnk->prev;
	if(chnk)pos=chnk->count-1;
      }else{
	--pos;
      }
    }
    void operator++(){//increment
      ++pos;
      if(pos==chnk->count){
	pos=0;
	chnk=chnk->next;
      }
    }
    mytype& operator*(){//dereference
      return chnk->data[pos];
    }
    int ready(){
      return chnk!=0;
    }
  };


 private:
  size_t list_size;
  size_t chunkcount;//not necessary for basics, but useful for diagnostics
  chunk* chunk_first;
  chunk* chunk_last;
  iterator first;
  iterator last;

  void expand();

 public:

  void allocate(size_t n);//make a new chunk and put it in the chunk list

  unroll();//empty
  unroll(size_t size);//initialize with size
  //O(1)+alloc
  void push_back(const mytype& val);
  //O(1)
  void pop_back();
  //O(1)
  mytype& back();
  mytype& front();

  void erase(uint i);

  iterator begin(){//inline
    return first;
  }
  iterator end(){//inline
    return last;
  }
  iterator get(uint i);

  //O(m)
  void append(unroll& ul);//transfers ownership

  void transfer(unroll& ul);//transfers ownership

  void head(size_t count);
  void tail(size_t count);

  void operator=(const unroll& ul);//creates clean copy
  unroll(const unroll& ul);//creates clean copy

  //O(m)
  mytype& operator[](int i);//somewhat clumsy
  //O(n)
  void reverse();//swaps elements
  //O(m)
  void clear();
  void straighten();
  //O(n log(n))
  void sort(int (*compare)(const mytype&, const mytype&));//sort
  mytype min(int (*compare)(const mytype&, const mytype&));//min element
  mytype accum(const mytype& start, void (mytype::*acc)(const mytype&));

  void map(void (mytype::*fun)());
  void filter(int (mytype::*fun)() const, unroll& out);
  int includes(const mytype& v);

  size_t size(){//inline
    return list_size;
  }
  ~unroll();//cleanup

  void components(unroll& ul1, unroll& ul2, mytype (mytype::*op)(const mytype&) const);
  void tensor(unroll& ul1, unroll& ul2, mytype (mytype::*op)(const mytype&) const);

  int operator==(const unroll& ul2);

#if DEBUG>1
  void print(FILE* fd, const char* Fd, const char* Ff);
  void print_short(FILE* fd);
#endif
};


typedef unroll array;

#endif
