/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2012  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "base.h"
#include "vm.h"
#include "methods.h"

#ifdef TIMERAND
#include <ctime>
#endif

#ifndef NOSYSCALL
#include <sys/times.h>
#include <unistd.h>
#endif

#include <list>
#include <algorithm>
#include <cstring>
#include <cctype>

//if no history, no storing history
#ifndef HAVE_READLINE_HISTORY
#undef STOREHISTORY
#endif

#ifdef STOREHISTORY
#include<sys/stat.h>
#endif

using namespace std;

#ifndef NOSYSCALL
void eval_control::syscall(array& a, array& out){
  int n=a.size();
  if(n==0){
#if DEBUG>0
    fprintf(stderr,"warning: empty syscall\n");
#endif
    return;
  }
  int syscl=a[0].assertI("Q{n ...}");
  switch(syscl){
    case SYSC_NOOP: {
      break;
    }
    case SYSC_EXIT: {
      shared::special_interrupt=INTERRUPT_EXIT;
      shared::exit_returnvalue=n==1?0:a[1].assertI("Q{1 n}");
      if(shared::exit_returnvalue&~0xff)shared::exit_returnvalue=1;
      break;
    }
    case SYSC_FLOW_BRK: {
      if(n==1 || (a[1]!=VAL_ZERO)){
	shared::special_interrupt=INTERRUPT_BREAK;
      }
      break;
    }
    case SYSC_FLOW_CONT: {
      if(n==1 || (a[1]!=VAL_ZERO)){
	shared::special_interrupt=INTERRUPT_CONT;
      }
      break;
    }
    case SYSC_FLOW_PC: {
      if(!functrl)return;
      out.push_back((itype)(*functrl));
      break;
    }
    case SYSC_FLOW_PCSET: {
      if(!functrl)return;
      int n=a.size();
      if(n==0)return;
      *functrl=a[1].assertI("Q{103 n}")-1;
      break;
    }
    case SYSC_FLOW_STCKLVL: {
      out.push_back((itype)argstack.size());
      break;
    }
    case SYSC_FLOW_MAXIT: {
      int n=a.size();
      if(n==1){
	out.push_back((itype)maxit);
      }else{
	int tmp=a[1].assertI("Q{105 n}");
	if(tmp>=0)maxit=tmp;
      }
      break;
    }
    case SYSC_HIST_COUNT: {
      if(toggle_history){
	out.push_back((itype)histsize);
      }else{
	out.push_back(VAL_NEG);
      }
      break;
    }
    case SYSC_HIST_ENABLE: {
      if(n<2){
	toggle_history=0;
      }else{
	toggle_history=a[1].assertI("Q{1001 n}")!=0;
      }
      break;
    }
    case SYSC_HIST_CLEAR: {
      for(int i=0;i<histsize;i++){
	memory.erase(memory.find(-1-i));
      }
      histsize=0;
      break;
    }
    case SYSC_MEM_LCPOS: {
      if(n>1){
	int newmemroot=a[1].assertI("Q{2000 n}");
	if(newmemroot<0){
	  fprintf(stderr,"runtime: index memory slot (%d) should be nonnegative\n",newmemroot);
	}else{
	  memroot=newmemroot;
	}
      }else{
	out.push_back((itype)memroot);
      }
      break;
    }
    case SYSC_MEM_LPOS: {
      if(n>1){
	int newmemdefault=a[1].assertI("Q{2001 n}");
	if(newmemdefault<0){
	  fprintf(stderr,"runtime: default memory slot (%d) should be nonnegative\n",newmemdefault);
	}else{
	  memdefault=newmemdefault;
	}
      }else{
	out.push_back((itype)memdefault);
      }
      break;
    }
    case SYSC_MEM_AOFF: {
      if(n>1){
	int newmempos=a[1].assertI("Q{2002 n}");
	if(newmempos<0){
	  fprintf(stderr,"runtime: current memory position (%d) should be nonnegative\n",newmempos);
	}else{
	  mempos=newmempos;
	}
      }else{
	out.push_back((itype)mempos);
      }
      break;
    }
    case SYSC_MEM_CLEAR: {
      if(n==1){
	for(t_mem::iterator i=memory.begin();i!=memory.end();){
	  if((*i).first<0){
	    i++;
	    continue;
	  }
	  i++;
	  memory.erase(i);
	}
      }else{
	for(int i=1;i<n;i++){
	  int indx=a[i].assertI("Q{2003 n}");
	  t_mem::iterator it=memory.find(indx);
	  if(it!=memory.end()){
	    memory.erase(it);
	  }else{
	    out.push_back((itype)indx);
	  }
	}
      }
      break;
    }
    case SYSC_MEM_LIST: {
      for(t_mem::iterator i=memory.begin();i!=memory.end();i++){
	if((*i).first<0)continue;
	out.push_back((itype)(*i).first);
      }
      break;
    }
    case SYSC_FUN_CLEAR: {
      if(n==1){
	funspace.clear();
      }else{
	for(int i=1;i<n;i++){
	  int indx=a[i].assertI("Q{2050 n}");
	  t_fun::iterator it=funspace.find(indx);
	  if(it!=funspace.end()){
	    funspace.erase(it);
	  }else{
	    out.push_back((itype)indx);
	  }
	}
      }
      break;
    }
    case SYSC_FUN_LIST: {
      for(t_fun::iterator i=funspace.begin();i!=funspace.end();i++){
	if((*i).first<0)continue;
	out.push_back((itype)(*i).first);
      }
      break;
    }
    case SYSC_BLD_VERSION: {
      char* ver=strdup(VERSION);
      char* v1=ver;
      char* v2=strchr(v1,'.');
      (*v2)=0;
      v2++;
      char* v3=strchr(v2,'.');
      (*v3)=0;
      v3++;
      out.push_back(value(v1));
      out.push_back(value(v2));
      out.push_back(value(v3));
      free(ver);
      break;
    }
    case SYSC_BLD_SINT: {
      out.push_back((itype)sizeof(itype));
      break;
    }
    case SYSC_BLD_SFLT: {
      out.push_back((itype)sizeof(ieee));
      break;
    }
    case SYSC_BLD_DEBUG: {
      out.push_back((itype)DEBUG);
      break;
    }
    case SYSC_BLD_RL: {
#ifdef HAVE_LIBREADLINE
      out.push_back(VAL_ONE);
#else
      out.push_back(VAL_ZERO);
#endif
#ifdef HAVE_READLINE_HISTORY
      out.push_back(VAL_ONE);
#else
      out.push_back(VAL_ZERO);
#endif
#ifdef STOREHISTORY
      out.push_back(VAL_ONE);
#else
      out.push_back(VAL_ZERO);
#endif
      break;
    }
    case SYSC_BLD_OVERFLOW: {
#ifdef TRAP_OVR
      out.push_back(VAL_ONE);
#else
      out.push_back(VAL_ZERO);
#endif
      break;
    }
    case SYSC_SYS_TIME: {
      time_t timer;
      time(&timer);
      tm* time_converted=localtime(&timer);
      if(n==1){
	out.push_back((itype)timer);
      }else{
	for(int i=1;i<n;i++){
	  int t_req = a[i].assertI("Q{5000 n}");
	  switch(t_req){
	    case SYSCTIME_EPOCH:
	      out.push_back((itype)timer);
	      break;
	    case SYSCTIME_SEC:
	      out.push_back((itype)time_converted->tm_sec);
	      break;
	    case SYSCTIME_MIN:
	      out.push_back((itype)time_converted->tm_min);
	      break;
	    case SYSCTIME_HOUR:
	      out.push_back((itype)time_converted->tm_hour);
	      break;
	    case SYSCTIME_MDAY:
	      out.push_back((itype)time_converted->tm_mday);
	      break;
	    case SYSCTIME_MON:
	      out.push_back((itype)(time_converted->tm_mon+1));
	      break;
	    case SYSCTIME_YEAR:
	      out.push_back((itype)(1900+time_converted->tm_year));
	      break;
	    case SYSCTIME_WDAY:{
	      itype wdy=time_converted->tm_wday;
	      if(wdy==0)wdy=7;
	      out.push_back(wdy);
	      break;
	    }
	    case SYSCTIME_YDAY:
	      out.push_back((itype)(time_converted->tm_yday+1));
	      break;
	    default:
	      fprintf(stderr,"runtime: (%d) is not valid time request\n",t_req);
	      break;
	  }
	}
      }
      break;
    }
    case SYSC_SYS_PTIME: {
      int tick=sysconf(_SC_CLK_TCK);
      tms buffer;
      clock_t rtime=times(&buffer)-start_time;
      out.push_back((ieee)(((ieee)rtime)/tick));
      out.push_back((ieee)(((ieee)buffer.tms_utime)/tick));
      out.push_back((ieee)(((ieee)buffer.tms_stime)/tick));
      break;
    }
    case SYSC_SYS_SLEEP: {
      for(int i=1;i<n;i++){
	value v=a[i];
	int sec=v.intpart();
	int nsec=(1000000000*(v.d-sec));
	if(nsec<0)nsec=0;
	if(nsec>1000000000)nsec=0;
	timespec t1={sec,nsec};
	timespec t2;
	nanosleep(&t1,&t2);
      }
      break;
    }
    case SYSC_SYS_PID: {
      out.push_back((itype)getpid());
      break;
    }
    default: {
#if DEBUG>0
      fprintf(stderr,"warning: unknown syscall (%d)\n",syscl);
#endif
      break;
    }
  }
}
#else
void eval_control::syscall(array& a, array& out){
  fprintf(stderr,"warning: operator 'Q' disabled at compile time\n");
}
#endif

//functions

#define a  args[0]
#define a1 args[0]
#define a2 args[1]

void eval_control::evaluate_operator(uint code, int argc, array* args, array& out){
  //static array of function pointers
  static value (value::*binops[8])(const value&) const={
    &value::operator+,
    &value::operator-,
    &value::operator*,
    &value::operator/,
    &value::operator%,
    &value::operator^,
    &value::e,
    &value::binomial 
  };
  static value (value::*logops[6])(const value&) const={
    &value::lt,
    &value::gt,
    &value::le,
    &value::ge,
    &value::eq,
    &value::neq
  };
  static void (value::*funmaps[18])()={
    &value::op_fact,
    &value::op_sin,
    &value::op_cos,
    &value::op_tan,
    &value::op_asin,
    &value::op_acos,
    &value::op_atan,
    &value::op_sqrt,
    &value::op_cbrt,
    &value::op_exp,
    &value::op_log,
    &value::op_neg,
    &value::op_abs,
    &value::op_round,
    &value::op_frac,
    &value::op_int,
    &value::op_float,
    &value::op_rand
  };

  static int (value::*funfilters[7])() const={
    &value::is_positive,
    &value::is_negative,
    &value::is_odd,
    &value::is_even,
    &value::is_int,
    &value::is_float,
    &value::is_prime
  };


  switch(code&OPMASK){
    case OPX_MEM:{
      switch(code){
	case OP_NONE: return;
	case OP_TAKE: {
	  int n=a.size();
	  for(int i=0;i<n;i++){
	    array* mem=0;
	    int index=a[i].assertI("${n}");
	    if(index<0){
	      if(-index<=histsize){
		mem=&memory[-histsize-1-index];
	      }else continue;
	    }else{
	      mem=&memory[index];
	    }
	    //copy because memory must survive
	    array memcopy(*mem);
	    out.append(memcopy);
	  }
	  return;
	}
	case OP_PUT:{
	  int n1=a1.size();
	  for(int i=0;i<n1;i++){
	    int index=a1[i].assertI("@{n}{...}");
	    if(index>=0)memory[index]=a2;
	  }
	  out.transfer(a2);
	  return;
	}
	case OP_SYSCALL:{
	  syscall(a,out);
	  return;
	}
	case OP_FCALL:{
	  int n1=a1.size();
	  out.transfer(a2);
	  if(n1==0)return;
	  int* oldfntctrl=functrl;//store old pointer, even if it is zero
	  for(int i=0;i<n1 && i>=0;i++){
	    int findex=a1[i].assertI("{n}::{...}");
	    if(findex==0)continue;//null function is no-op
	    if(funspace.find(findex)==funspace.end()){
	      fprintf(stderr,"lambda: invalid function id (%d)\n",findex);
	      continue;
	    }
	    treenode* tn=&funspace[findex];
	    argstack.push(out);
	    out.clear();
	    functrl=&i;//pass the reference to i
	    int evalret=evaluate(*tn,out);
	    argstack.pop();
	    if(evalret)break;
#ifdef SIGHANDLE
	    if(shared::interrupt){break;}
#endif
	  }
	  functrl=oldfntctrl;//restore the pointer
	  return;
	}
	case OP_FITER:{
	  int n1=a1.size();
	  if(n1==0)return;

	  int* oldfntctrl=functrl;//store old pointer, even if it is zero

	  array previous;
	  array current=a2;

	  int unfinished=1;

	  //check if everything is ok and get function indices
	  int* funindices=(int*)malloc(n1);
	  for(int i=0;i<n1;i++){
	    int findex=a1[i].assertI("{n}:::{...}");
	    if(funspace.find(findex)==funspace.end()){
	      fprintf(stderr,"lambda: invalid function id (%d), ignoring\n",findex);
	      findex=0;
	    }
	    funindices[i]=findex;
	  }

	  int loopindex;
	  for(loopindex=0;loopindex<maxit && unfinished;++loopindex){
	    previous=current;
	    for(int i=0;i<n1 && i>=0;i++){
	      int findex=funindices[i];
	      if(findex==0)continue;//null function is no-op
	      treenode* tn=&funspace[findex];
	      argstack.push(current);
	      current.clear();
	      functrl=&i;//pass the reference to i
	      int evalret=evaluate(*tn,current);
	      argstack.pop();
	      if(evalret)break;
#ifdef SIGHANDLE
	      if(shared::interrupt){break;}
#endif
	    }
#ifdef SIGHANDLE
	    if(shared::interrupt){break;}
#endif
	    if(previous==current)unfinished=0;
	  }
	  if(loopindex==maxit){
	    fprintf(stderr,"warning: number of iterations exceeded maximum (%d)\n",maxit);
	  }
	  functrl=oldfntctrl;//restore the pointer

	  out.transfer(current);
	  free(funindices);
	  return;
	}
	case OP_X:{
	  if(argstack.empty()){
	    fprintf(stderr,"lambda: 'x' called outside a function\n");
	    return;
	  }
	  out=argstack.top();
	  return;
	}

      }
      break;
    }
    case OPX_COM:{
      out.components(a1,a2,binops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_ACC:{
      switch(code){
	case OP_OPLUS:{
	  out.push_back(a.accum(VAL_ZERO,&value::operator+=));
	  return;
	}
	case OP_OTIMES:{
	  out.push_back(a.accum(VAL_ONE,&value::operator*=));
	  return;
	}
	case OP_OMEAN:{
	  value v=a.accum(VAL_FZERO,&value::operator+=);
	  v/=(itype)a.size();
	  out.push_back(v);
	  return;
	}
	case OP_OVAR:{
	  int n=a.size();
	  value v=VAL_FZERO;
	  value v2=VAL_FZERO;

	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it);
	    v2+=(*it)*(*it);
	  }
	  v/=(itype)n;
	  v2/=(itype)n;
	  v*=v;
	  out.push_back(v2-v);

	  return;
	}
	case OP_ONORM:{
	  int n1=a1.size();
	  out.allocate(n1);
	  a2.map(&value::op_abs);
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    value v=VAL_ZERO;
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      v+=((*it2)^v1);
	    }
	    out.push_back(v^(ieee)(1/v1.d));
	  }
	  return;
	}
      }
      break;
    }
    case OPX_FUN:{
      out.transfer(a);
      out.map(funmaps[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_FUNX:{
      int n1=a1.size();
      int n2=a2.size();
      out.allocate(n1*n2);

      switch(code){
	case OP_BESJ:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int besn=(*it1).assertI("bj{n}{...}");
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      out.push_back((*it2).op_bj(besn));
	    }
	  }
	  return;
	}
	case OP_BESY:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int besn=(*it1).assertI("by{n}{...}");
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      out.push_back((*it2).op_by(besn));
	    }
	  }
	  return;
	}
      }
      break;
    }
    case OPX_TEN:{
      out.tensor(a1,a2,binops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_SET:{
      switch(code){
	case OP_AND:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(a2.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_WITHOUT:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(!a2.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_OR:{
	  out.append(a1);
	  out.append(a2);
	  return;
	}
	case OP_XOR:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    value v1=(*it1);
	    if(!a2.includes(v1))out.push_back(v1);
	  }
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    value v2=(*it2);
	    if(!a1.includes(v2))out.push_back(v2);
	  }
	  return;
	}
	case OP_REVERSE:{
	  out.transfer(a);
	  out.reverse();
	  return;
	}
	case OP_SORT:{
	  out.transfer(a);
	  out.sort(operator<);
	  return;
	}
	case OP_HASH:{
	  out.push_back((itype)a.size());
	  return;
	}
	case OP_MASK:{
	  int n=a.size();
	  array::iterator it=a.begin();
	  for(int i=0;i<n;i++){
	    if((*it)!=VAL_ZERO){
	      out.push_back((itype)i);
	    }
	    ++it;
	  }
	  return;
	}
	case OP_EQUALS:{
	  int i=a1==a2?1:0;
	  out.push_back((itype)i);
	  return;
	}
	case OP_COUNT:{
	  for(array::iterator it1=a1.begin();it1.ready();++it1){
	    int count=0;
	    value x=(*it1);
	    for(array::iterator it2=a2.begin();it2.ready();++it2){
	      if(x==(*it2))count++;
	    }
	    out.push_back((itype)count);
	  }
	  return;
	}
	case OP_HIST:{
	  int n=a1.size();
	  out.allocate(n+1);
	  //initialize counts; bins are (-inf,a0),(a0,a1),...,(an-1,inf)
	  for(int i=0;i<=n;i++){
	    out.push_back(VAL_ZERO);
	  }
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int index=0;
	    value x=(*it2);
	    for(array::iterator it1=a1.begin();it1.ready();++it1){
	      if(x<(*it1))break;
	      index++;
	    }
	    out[index]+=VAL_ONE;
	  }
	  return;
	}

      }
      break;
    }
    case OPX_SEL:{
      switch(code){
	case OP_SHL:{
	  int n1=a1.size();
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int ihi=(*it2).assertI("{...}<<{n}");
	    if(ihi<0){//negative head is tail
	      ihi=-ihi;
	      for(int i=n1;i<ihi;i++)out.push_back(VAL_ZERO);//padding
	      array copy=a1;
	      copy.tail(n1-ihi);
	      out.append(copy);
	    }else{
	      array copy=a1;
	      copy.head(ihi);
	      out.append(copy);
	      for(int i=n1;i<ihi;i++)out.push_back(VAL_ZERO);//padding
	    }
	  }
	  return;
	}
	case OP_SHR:{
	  int n1=a1.size();
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    int ihi=(*it2).assertI("{...}>>{n}");
	    if(ihi<0){//negative tail is head
	      ihi=-ihi;
	      array copy=a1;
	      copy.head(n1-ihi);
	      out.append(copy);
	      for(int i=n1;i<ihi;i++)out.push_back(VAL_ZERO);//padding
	    }else{
	      for(int i=n1;i<ihi;i++)out.push_back(VAL_ZERO);//padding
	      array copy=a1;
	      copy.tail(ihi);
	      out.append(copy);
	    }
	  }
	  return;
	}
	case OP_INDEX:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(!n1 || !n2)return;

	  out.allocate(n2);
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    value v=(*it2);//%(itype)n1;
	    if(v.type==INTEGER){
	      out.push_back(a1[v.i]);
	    }else{
	      itype ii=v.intpart();
	      ieee f=v.d-ii;
	      if(f<0 || f>1){
		fprintf(stderr,"runtime: warning, array subscript (%g) out of integer range\n",v.d);
		continue;
	      }
	      out.push_back(a1[ii]*(1-f)+a1[ii+1]*f);
	    }
	  }
	  return;
	}
	case OP_UNIQUE:{
	  for(array::iterator it=a.begin();it.ready();++it){
	    value v1=(*it);
	    if(!out.includes(v1))out.push_back(v1);
	  }
	  return;
	}
	case OP_MIN:{
	  if(!a.size())return;
	  out.push_back(a.min(operator<));
	  return;
	}
	case OP_MAX:{
	  if(!a.size())return;
	  out.push_back(a.min(operator>));
	  return;
	}

        break;
      }
    }
    case OPX_FILT:{
      a.filter(funfilters[(code&(~OPMASK))-1],out);
      break;
    }
    case OPX_CLOG:{
      out.components(a1,a2,logops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_TLOG:{
      out.tensor(a1,a2,logops[(code&(~OPMASK))-1]);
      break;
    }
    case OPX_CONST:{
      switch(code){
	case OP_PI:{
	  out.push_back(VAL_PI);
	  return;
	}
	case OP_AA:{
	  out.push_back(VAL_AA);
	  return;
	}
	case OP_DEG:{
	  out.push_back(VAL_DEG);
	  return;
	}
	case OP_EMC:{
	  out.push_back(VAL_EMC);
	  return;
	}
	case OP_PM:{
	  out.push_back(VAL_ONE);
	  out.push_back(VAL_NEG);
	  return;
	}
	case OP_MP:{
	  out.push_back(VAL_NEG);
	  out.push_back(VAL_ONE);
	  return;
	}
 
      }
      break;
    }
    case OPX_POLY:{
      switch(code){
	case OP_XPLUS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(n1<n2){
	    out.transfer(a2);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a1.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }else{
	    out.transfer(a1);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a2.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }
	  return;
	}
	case OP_XMINUS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(n1<n2){
	    out.transfer(a2);
	    out.map(&value::op_neg);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a1.begin();it.ready();++it){
	      (*ito)+=(*it);
	      ++ito;
	    }
	  }else{
	    out.transfer(a1);
	    array::iterator ito=out.begin();
	    for(array::iterator it=a2.begin();it.ready();++it){
	      (*ito)-=(*it);
	      ++ito;
	    }
	  }
	  return;
	}
	case OP_XMUL:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1+n2-1;
	  out.allocate(n);
	  for(int i=0;i<n;i++){
	    value v=VAL_ZERO;
	    for(int j=0;j<=i;j++){
	      if((j<n1)&&(i-j<n2)){
		v+=a1[j]*a2[i-j];
	      }
	    }
	    out.push_back(v);
	  }
	  return;
	}
	case OP_XDIV:{
	  trimzeroes(a1);
	  trimzeroes(a2);
	  int n1=a1.size();
	  int n2=a2.size();
	  if((n1==0) || (n2==0)){
	    return;
	  }
	  int n=polydivmod(a1,a2,out);
	  out.tail(out.size()-n);
	  if(!out.size())out.push_back(VAL_ZERO);
	  return;
	}
	case OP_XMOD:{
	  trimzeroes(a1);
	  trimzeroes(a2);
	  int n1=a1.size();
	  int n2=a2.size();
	  if((n1==0) || (n2==0)){
	    return;
	  }
	  int n=polydivmod(a1,a2,out);
	  out.head(n);
	  if(!out.size())out.push_back(VAL_ZERO);
	  return;
	}

	case OP_PX:{
	  int n1=a1.size();
	  int n2=a2.size();
	  if(!n1)return;//fixed bug
	  out.allocate(n2);
	  for(array::iterator it2=a2.begin();it2.ready();++it2){
	    if(n1==1){
	      out.push_back(a1.front());
	      continue;
	    }

	    value x=(*it2);
	    if(fabs(x.d)<1 || n1<3 || x.is_int()){
	      value v=a1[n1-1];
	      for(int j=n1-2;j>=0;j--){
		v*=x;
		v+=a1[j];
	      }
	      out.push_back(v);
	    }else{
	      value v=a1.front();
	      for(int j=1;j<n1;j++){
		v/=x;
		v+=a1[j];
	      }
	      v*=x^((itype)(n1-1));//fixed n2->n1!!!
	      out.push_back(v);
	    }
	  }
	  return;
	}
	case OP_PD:{
	  int n=a.size();
	  if(n<=1){
	    out.push_back(VAL_ZERO);
	    return;
	  }
	  out.allocate(n-1);
	  array::iterator it=a.begin();
	  ++it;
	  for(int i=1;i<n;i++){
	    out.push_back((*it)*(itype)i);
	    ++it;
	  }
	  return;
	}
	case OP_PZ:{
	  bairstow(a,out);
	  return;
	}

      }
      break;
    }
    case OPX_NUMBR:{
      switch(code){
	case OP_GCD:{
	  gcd(a,out);
	  break;
	}
	case OP_LCM:{
	  lcm(a,out);
	  break;
	}
	case OP_FACTOR:{
	  for(array::iterator it=a.begin();it.ready();++it){
	    if((*it).type!=INTEGER){
	      out.push_back(*it);
	      continue;
	    }
	    pfactor((*it).i,out);
	  }
	  break;
	}
	case OP_EFI:{
	  for(array::iterator it=a.begin();it.ready();++it){
	    if((*it).type==INTEGER){
	      if((*it).i<1)continue;
	      out.push_back(onephi((*it).i));
	    }
	  }
	  break;
	}
      }
      break;
    }
    case OPX_VEC:{
      switch(code){
	case OP_DOT:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1<n2?n1:n2;
	  value v=VAL_ZERO;
	  array::iterator it1=a1.begin();
	  array::iterator it2=a2.begin();
	  for(int i=0;i<n;i++){
	    v+=(*it1)*(*it2);
	    ++it1;
	    ++it2;
	  }
	  out.push_back(v);
	  return;
	}
	case OP_VLEN:{
	  int n=a.size();
	  value v=VAL_ZERO;
	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it)*(*it);
	  }
	  v.op_sqrt();
	  out.push_back(v);
	  return;
	}
	case OP_VNRM:{
	  int n=a.size();
	  value v=VAL_ZERO;
	  for(array::iterator it=a.begin();it.ready();++it){
	    v+=(*it)*(*it);
	  }
	  v.op_sqrt();
	  for(array::iterator it=a.begin();it.ready();++it){
	    (*it)/=v;
	  }
	  out.transfer(a);
	  return;
	}
	case OP_CROSS:{
	  int n1=a1.size();
	  int n2=a2.size();
	  int n=n1<n2?n1:n2;
	  out.allocate(n);
	  if(n==1){out.push_back(VAL_ZERO);return;}
	  for(int i=0;i<n;i++){
	    out.push_back(a1[i+1]*a2[i+2]-a1[i+2]*a2[i+1]);//automatic index wrapping
	  }
	  return;
	}

      }
      break;
    }
    case OPX_PAREN:{
      switch(code){

	case OP_LIST:{
	  for(int i=0;i<argc;i++){
	    out.append(args[i]);
	  }
	  return;
	}
	case OP_ITER1:{
	  array arg;
	  for(int i=0;i<argc;i++){
	    arg.append(args[i]);
	  }
	  int n=arg.size();
	  switch(n){
	    case 0:{
	      out.push_back((itype)42);
	      break;
	    }
	    case 1:{
	      int count=arg.front().assertI("[n]");
	      out.allocate(count);
	      for(int i=0;i<count;i++){
		out.push_back((itype)i);
	      }
	      break;
	    }
	    case 2:{
	      value v0=arg.front();
	      value v1=arg.back();
	      if(v1>v0){
		for(value v=v0;v.d<=v1.d;v+=VAL_ONE){
		  out.push_back(v);
		}
	      }else{
		for(value v=v0;v.d>=v1.d;v-=VAL_ONE){
		  out.push_back(v);
		}
	      }
	      break;
	    }
	    default:{
	      value v=arg.front();
	      value v1=arg.back();

	      //same value?

	      if(v.d==v1.d){
		out.push_back(v);
		break;
	      }
	      value max=VAL_ZERO; //max. offset
	      value min=VAL_ZERO;
	      value sum=VAL_ZERO;
	      for(int i=1;i<n-1;i++){
		sum+=arg[i];
		if(sum>max)max=sum;
		if(sum<min)min=sum;
	      }
	      //is there minus?
	      if(v>v1){//if the sequence is descending or if it overshot final value
		if((sum.d<0)||(min<v1-v)){
		  int index=0;
		  for(int i=0;v.d>=v1.d;i++){
		    out.push_back(v);
		    v+=arg[1+index];
		    index++;
		    if(index>=(n-2))index-=(n-2);
		  }
		}
	      }else{//if the sequence is ascending or if it overshot final value
		if((sum.d>0)||(max>v1-v)){
		  int index=0;
		  for(int i=0;v.d<=v1.d;i++){
		    out.push_back(v);
		    v+=arg[1+index];
		    index++;
		    if(index>=(n-2))index-=(n-2);
		  }
		}
	      }
	      break;
	    }
	  }
	  return;
	}
	case OP_ITER2:{
	  array arg;
	  for(int i=0;i<argc;i++){
	    arg.append(args[i]);
	  }
	  int n=arg.size();
	  switch(n){
	    case 0:break;
	    case 1:{
	      out.push_back(arg.front());
	      break;
	    }
	    case 2:{
	      ieee limit=arg.back().d;
	      if(limit<0)break;
	      out.allocate((int)(limit)+1);
	      value v=arg.front();
	      for(int i=0;i<limit;i++){
		out.push_back(v);
	      }
	      break;
	    }
	    default:{
	      value v=arg.front();
	      ieee count=arg.back().d;//arg[n-1].d;
	      if(count<0)count=(-count)*(n-2)+1;//negative: whole sequence times count
	      out.allocate((int)(count)+1);
	      int index=0;
	      for(int i=0;i<count;i++){
		out.push_back(v);
		v+=arg[1+index];
		index++;
		if(index>=(n-2))index-=(n-2);
	      }
	      break;
	    }
	  }
	  return;
	}

      }
      break;
    }
      //end of outer switch
  }
}

#undef a
#undef a1
#undef a2

int eval_control::evaluate(treenode& node, array& out){
  switch(node.type){
    case NODE_OP:{
      //function operator is very very special!
      if(node.opcode==OP_FUN){
	out.push_back((itype)funcount);
	funspace[funcount]=node;
	funspace[funcount].opcode=shared::shared_op_parL.code;
	funcount++;
	return 0;
      }

      int argc=node.childz.size();
      array* args=new array[argc];
      int count=0;
      for(list<treenode>::iterator i=node.childz.begin();i!=node.childz.end();i++){
#ifdef SIGHANDLE
        //interrupt
        if(shared::interrupt){
	  delete[] args;
	  return 1;
	}
#endif
	if(evaluate(*i,args[count])){//recursion!!!
	  delete[] args;
	  return 1;
	}
	count++;
      }
      evaluate_operator(node.opcode,argc,args,out);
      //out.print(stderr,Fd,Ff);
      out.straighten();
      delete[] args;
#ifndef NOSYSCALL
      if(shared::special_interrupt){
	return 1;//interrupt if exit syscall issued
      }
#endif
      return 0;
    }
    case NODE_VAL:{
      out.push_back(node.val);
      return 0;
    }
  }
}


void eval_control::print_array(array& a, FILE* stream, const char* sep, char end){
    if(!a.size()){if(end)fputc(end,stream);return;}
    int count=a.size()-1;
    array::iterator it=a.begin();
    for(int i=0;i<count;i++){
      (*it).print(stream,Fd,Ff);
      ++it;
      if(sep)fputs(sep,stream);
#ifdef SIGHANDLE
      //interrupt - long outputs take forever, must be able to break loop
      if(shared::interrupt){fputc('\n',stream);return;}
#endif
    }
    a.back().print(stream,Fd,Ff);
    if(end)fputc(end,stream);
}

char* eval_control::strip(char* str){
  char* out=(char*)malloc(strlen(str)+1);
  char* pos=out;//current character
  char* end=out;//position of null
  char mode=0; //quote type
  for(;*str;str++){
    if(!mode){
      if(isspace(*str))continue;
      if(*str=='\''){mode='\'';continue;}
      if(*str=='"'){mode='"';continue;}
      if(*str=='\\'){mode=' ';str--;continue;}//we handle escapes later
      mode=' ';
      *pos=*str;pos++;end=pos;
      continue;
    }
    //escaped characters (not for '')
    if((*str=='\\')&&(mode!='\'')){
      if(!str[1]);else
      if(str[1]=='t'){*pos='\t';pos++;}else
      if(str[1]=='n'){*pos='\n';pos++;}else
      if(str[1]=='r'){*pos='\r';pos++;}else
      if(str[1]=='f'){*pos='\f';pos++;}else
      if(str[1]=='b'){*pos='\b';pos++;}else
      if(str[1]=='v'){*pos='\v';pos++;}else
	{*pos=str[1];pos++;}

      /*
      switch(str[1]){
        case 0: break;
        case 't': *pos='\t';pos++;break;
        case 'n': *pos='\n';pos++;break;
        case 'r': *pos='\r';pos++;break;
        case 'f': *pos='\f';pos++;break;
        case 'b': *pos='\b';pos++;break;
        case 'v': *pos='\v';pos++;break;
        default: *pos=str[1];pos++;
      }
      */
      //escaped characters count even at the end
      end=pos;
      if(str[1]){
        str++;
      }else{//null after escape
        fprintf(stderr,"input: escape character at the end of input\n");
      }
      continue;
    }
    if((mode!=' ')&&(*str==mode)){
      mode=' ';//reset mode
      break;
    }
    *pos=*str;pos++;
    //if no quotes, spaces at the end don't count
    if(!((mode==' ')&&isspace(*str)))end=pos;
  }
  if(!mode){//etirely empty line
    free(out);
    return 0;
  }
  if(mode!=' '){//there was no closing quote
   fprintf(stderr,"input: missing closing quote '%c'\n",mode);
  }
  *end=0;//put the null where the string ends
  return out;
}

FILE* eval_control::input_file_open(char* filename){
  if(!strcmp(filename,"-"))return stdin;
  FILE* f=fopen(filename,"r");
  if(f)return f;
  fprintf(stderr,"input: file '%s' could not be opened\n",filename);
  return 0;
}

FILE* eval_control::output_file_open(char* filename){
  if(!strcmp(filename,"-"))return stdout;
  FILE* f=fopen(filename,"w");
  if(f)return f;
  fprintf(stderr,"output: file '%s' could not be opened\n",filename);
  return 0;
}


void eval_control::evaluate_code(list<code_chunk>& code, bool printlast, array& out){
  int i=0;
  for(list<code_chunk>::iterator it=code.begin();it!=code.end();++it){
    array a;
    if(evaluate(it->code,a))return;
    if(i==code.size()-1 && it->printable){
      if(printlast){
	print_array(a,outstream,delim,'\n');
      }
      out.transfer(a);
      return;
    }
    //print if forced print is enabled or semicolon is missing
    if(it->printable || toggle_print)print_array(a,outstream,delim,'\n');
    i++;
  }
  return;
}


//no line comes out of precompile with newline in it!!!!
//returns the number of noncommented lines
int eval_control::precompile(char* line, char** comment,list<code_chunk>& code){
  //convert newline into a null
  char* newline=strchr(line,'\n');
  if(newline)*newline=0;

  //shabang lines should be ignored. This way you can call ised with
  //#!/usr/local/bin/ised --f
  char* shabang=strstr(line,"#!");
  //if line begins with #!, omit it.
  //if line is empty, omit it.
  if((shabang==line)||(*line==0)){if(comment)*comment=0;return 0;}
  //if there's no comment, comment is end of string.
  if(comment)*comment=shabang?shabang:strchr(line,0);
  if(shabang)*shabang=0;//premature end of line at comment

  char* end;
  while(end=strchr(line,';')){//semicolon separates chunks of code
#ifdef SIGHANDLE
    //interrupt
    if(shared::interrupt){if(comment)*comment=0;return 0;}
#endif
    *end=0;//fake end
    code.push_back((code_chunk){treenode(shared::shared_op_parL),false});
    compiler(line,code.back().code);
    *end=';';//restore end
    line=end+1;
  }
  //something remains
  if(*line){
    //finalization
    code.push_back((code_chunk){treenode(shared::shared_op_parL),true});
    compiler(line,code.back().code);
    if(shabang)*shabang='#';//restore original line
    return 1;
  }else{
    if(shabang)*shabang='#';//restore original line
    return 0;
  }
}

char* eval_control::formatdup(const char* fmt){
#if IVALUE==ILONG || VALUE==LONG
  char* out=(char*)malloc(strlen(fmt)+2);//it may only be one character longer
  char* out_chr=out;

  uint percent=0;
  for(const char* c=fmt;*c;c++){
    if(*c == '%'){
      //percent: 0 -> 1 (found percentage)
      //percent: 1 -> 0 (%% = single percentage as character)
      //percent: anything else -> anything else (if more than one format, don't do anything)
      percent^=0x1;
    }else{
      if(percent==1 && isalpha(*c)){//percent is closed by a specifier
#if IVALUE==ILONG
	if(strchr("dxXiu",*c)){*out_chr='l';out_chr++;}
#endif
#if IVALUE==ILONG && VALUE==LONG
	else
#endif
#if VALUE==LONG
	if(strchr("eEfgG",*c)){*out_chr='L';out_chr++;}
#endif
	//in case of l or L: already long specifier, no problem
	//if shouldn't be there, not our problem
	percent=2;//percent + alphanumeric means percent closed
      }
    }
    *out_chr=*c;
    out_chr++;
  }
  *out_chr=0;//null terminated string
  //fprintf(stderr,"format='%s'\n",out);
  return out;
#else
  return strdup(fmt);
#endif
}

char* eval_control::bufgets(FILE* file){
   char* out=0;
   //put some dummy non-null character
   buffer[bufsize-1]=' ';
   out=fgets(buffer,bufsize,file);
   //if it changed, the buffer was too short
   while(buffer[bufsize-1]!=' '){
     #if DEBUG>0
     fprintf(stderr,"input: buffer resize from %d to %d\n",bufsize,2*bufsize);
     #endif
     buffer=(char*)realloc(buffer,2*bufsize);
     buffer[2*bufsize-1]=' ';
     out=fgets(buffer+bufsize-1,bufsize+1,file);
     bufsize*=2;
   }
   return out;
}

#ifndef HAVE_LIBREADLINE
char* readline(const char* prompt){
  fputs(prompt,stdout);
  char* buf=(char*)malloc(BUF_SIZE);
  return fgets(buf,BUF_SIZE,stdin);
}
#endif


void eval_control::read_interactive(){
  credits_dump();
#ifdef STOREHISTORY
  char* home=getenv("HOME");
  char* hist_file=0;
  if(home){
    hist_file=(char*)malloc(strlen(home)+15);
    sprintf(hist_file,"%s/.ised_history",home);

    struct stat buf;
    if(!stat(hist_file,&buf)){//file exists
      if(int errno=read_history(hist_file)){
	fprintf(stderr,"readline: errno %d on history load\n",errno);
      }
    }
  }
#endif

#ifdef SIGHANDLE
  shared::wait_ptr=&awaiting;
#endif
  char prompt[7];
  strcpy(prompt,"ised> ");
  char* line;
  while(line=readline(prompt)){
#ifdef SIGHANDLE
    shared::state=STATE_WORK;
#endif
    //check for interactive interrupt
    if(!strcmp(line,"q")|| !strcmp(line,"quit"))break;
    input_line(line);
#ifndef NOSYSCALL
    if(shared::special_interrupt==INTERRUPT_EXIT)break;
#endif
#ifdef SIGHANDLE
    shared::state=STATE_IDLE;
    shared::interrupt=0;
#endif
#ifdef HAVE_READLINE_HISTORY
    if(*line)add_history(line);
#endif
    free(line);
    //change prompt:
    if(awaiting){
      sprintf(prompt,"%c= ",awaiting);
    }else{
      strcpy(prompt,"ised> ");
    }
  }
#ifdef SIGHANDLE
  shared::wait_ptr=0;
#endif

#ifdef STOREHISTORY
  if(home){
    if(int errno=write_history(hist_file)){
      fprintf(stderr,"readline: errno %d on history write\n",errno);
    }else{
      history_truncate_file(hist_file,256);
    }
  }
  if(hist_file)free(hist_file);
#endif

}


void eval_control::read_file(FILE* f){
  while(bufgets(f)){
    input_line(buffer);
#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif

  }
}


void eval_control::evaluate_line(char* line, char** comment, bool printlast, array& out){
  list<code_chunk> code;
  precompile(line,comment,code);
  out.clear();//not sure if necessary!
  evaluate_code(code,printlast,out);
}

void eval_control::load_file(FILE* f){
  char* comment;
  array a;
  while(bufgets(f)){
    evaluate_line(buffer,&comment,false,a);

#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif

    //save to memory unless it's commented or empty line
    if(comment)memory[++mempos].transfer(a);
  }
  memory[memroot].clear();
  memory[memroot].push_back((itype)mempos);
}

//returns 0 for normal code, 1 for nonprintable code and 2 for immediate return
int eval_control::retrieve_code(char* f, list<code_chunk>& code, bool file_or_line){
  char* comment;
  int outputlines=0;
  if(file_or_line){
    FILE* ffile=input_file_open(f);
    if(!ffile)return 2;
    while(bufgets(ffile)){
      outputlines=precompile(buffer,&comment,code);
    }

    fclose(ffile);
  }else{
    outputlines=precompile(f,&comment,code);
    if(!comment)return 2;//empty, failed or entirely commented
  }
  return outputlines==0?1:0;
}

void eval_control::apply_to_file(char* f, FILE* l, bool file_or_line){
  char* comment;

  list<code_chunk> code;
  int errcode=retrieve_code(f,code,file_or_line);
  if(errcode==2)return;
  int line=1;
  while(bufgets(l)){
    memory[memroot].clear();
    memory[memroot].push_back((itype)line);
    evaluate_line(buffer,&comment,false,memory[memdefault]);
#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif


    if(comment)line++;
    if(comment){
      array a;
      evaluate_code(code,false,a);
#ifndef NOSYSCALL
      if(shared::special_interrupt){
	if(shared::special_interrupt==INTERRUPT_BREAK){
	  shared::special_interrupt=INTERRUPT_NONE;break;
	}
	if(shared::special_interrupt==INTERRUPT_CONT){
	  shared::special_interrupt=INTERRUPT_NONE;continue;
	}
	return;
      }
#endif

      if(!errcode){//only print if there's something to print
        print_array(a,outstream,delim,0);
        if(*comment){
	  fputs(comment,outstream);
        }
	fputc('\n',outstream);
      }
    }else{//if empty, failed or entirely commented, preserve it (preserves blanks)
      if(!errcode){
	fputs(buffer,outstream);
	fputc('\n',outstream);
      }
    }
  }
}

void eval_control::interlace_code(list<code_chunk>& code){
  int codesize=code.size();
  array* temp_buffer=new array[codesize];
  int count=0;

  for(list<code_chunk>::iterator it=code.begin();it!=code.end();++it){
    if(evaluate(it->code,temp_buffer[count])){
      delete[] temp_buffer;
      return;
    }
    count++;
  }

  int index=0;
  while(1){
    array a;
    for(int j=0;j<codesize;j++){
      if(temp_buffer[j].size()>index){
        a.push_back(temp_buffer[j][index]);
      }
    }
    if(a.size()>0){
      print_array(a,outstream,delim,'\n');
      index++;
    }else break;
  }
  delete[] temp_buffer;
}

void eval_control::interlace_to_line(char* line){
  list<code_chunk> code;
  precompile(line,0,code);
  interlace_code(code);
}


void eval_control::interlace_to_file(char* f, FILE* l){
  list<code_chunk> code;
  precompile(f,0,code);

  char* comment;

  int line=1;
  while(bufgets(l)){
    memory[memroot].clear();
    memory[memroot].push_back((itype)line);
    evaluate_line(buffer,&comment,false,memory[memdefault]);
#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif
    if(comment){
      line++;
      interlace_code(code);
    }else{
      fputs(buffer,outstream);
      fputc('\n',outstream);
    }
#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif
  }
}

void eval_control::memory_dump(){
  typedef t_mem::iterator iterator;
  fprintf(stderr,"memory page:\n");
  for(iterator i=memory.begin();i!=memory.end();i++){
    if((*i).first<0)continue;
    fprintf(stderr,"@%+6d | ",(*i).first);
    print_array((*i).second,stderr," ",'\n');
  }
  fprintf(stderr,"end of memory page\n");
}

void eval_control::credits_dump(){
fprintf(stderr,"%s  Copyright (C) 2009-2012 Simon Copar\n\
This program comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it\n\
under certain conditions.\n",
PACKAGE_STRING);
}

static const char* STR_DISABLED="DISABLED";
static const char* STR_ENABLED="ENABLED";

void eval_control::configure_dump(){
#if VALUE==FLOAT
const char* dstr="FLOAT";
#elif VALUE==DOUBLE
const char* dstr="DOUBLE";
#elif VALUE==LONG
const char* dstr="LONG DOUBLE";
#endif

#if IVALUE==IINT
const char* istr="INT";
#elif IVALUE==ILONG
const char* istr="LONG LONG";
#endif
fprintf(stderr,"Build parameters:\n\
  floating point precision=%s (%d bytes)\n\
  integer precision=%s (%d bytes)\n",
  dstr,sizeof(ieee),istr,sizeof(itype));
#ifndef HAVE_LIBREADLINE
 const char* lrlstr=STR_DISABLED;
#else
 const char* lrlstr=STR_ENABLED;
#endif
#ifndef HAVE_READLINE_HISTORY
 const char* rlhstr=STR_DISABLED;
#else
 const char* rlhstr=STR_ENABLED;
#endif
#ifndef STOREHISTORY
 const char* rlhstore=STR_DISABLED;
#else
 const char* rlhstore=STR_ENABLED;
#endif
#ifndef TRAP_OVR
 const char* ovrflw=STR_DISABLED;
#else
 const char* ovrflw=STR_ENABLED;
#endif
 fprintf(stderr,"  libreadline %s, history %s, history store %s, overflow handling %s\n", lrlstr,rlhstr,rlhstore,ovrflw);

}

void eval_control::input_line(char* line){
#ifndef NOSYSCALL
  shared::special_interrupt=INTERRUPT_NONE;
#endif
  if(!strncmp(line,"#!",2))return;
  char* tofree=0;
  if(!awaiting){
    if(!strncmp(line,"--m",3)){memory_dump();return;}
    if(!strncmp(line,"--p",3)){toggle_print=1;return;}
    if(!strncmp(line,"--n",3)){toggle_print=0;return;}
    if(!strncmp(line,"--v",3)){
      credits_dump();
      putc('\n',stderr);
      configure_dump();
      return;
    }
    const char* chars="faldtDFxo";
    char compare[4]="---";

    while(*chars){
      compare[2]=*chars;
      if(!strncmp(line,compare,3)){
	tofree=line=strip(line+3);
        awaiting=*chars;
        break;
      }
      chars++;
    }
  }

  //waiting for something
  if(awaiting){
    if((!tofree)&&line){
      if(awaiting=='t'){
	tofree=line=strdup(line);
      }else{
	tofree=line=strip(line);
      }
    }
    if(!line)return;//really empty line
    switch(awaiting){
      case 'f':{
        awaiting=0;
	if(lfile){//if we want to apply to another file
	  FILE* l=input_file_open(lfile);
	  if(l){
	    if(xltoggle==XL_LINE){
	      apply_to_file(line,l,true);
	    }else{
	      extended_to_file(line,l,true);
	    }
	    fclose(l);
	  }
	  free(lfile);
	  lfile=0;
	}else{
	  FILE* f=input_file_open(line);
	  if(f){
	    read_file(f);//just read
	    fclose(f);
	  }
	}
        break;
      }
      case 'a':{
        awaiting=0;
        FILE* f=input_file_open(line);
        if(f){
          load_file(f);//load into memory
          fclose(f);
        }
        break;
      }
      case 'l':{
        awaiting=0;
	xltoggle=XL_LINE;
        if(lfile)free(lfile);
        lfile=strdup(line);
        break;
      }
      case 'o':{
        awaiting=0;
	//output stream
	fflush(outstream);
	if(outstream && outstream!=stdout)fclose(outstream);
	outstream=output_file_open(line);
	if(!outstream)outstream=stdout;
        break;
      }
      case 'x':{
	awaiting=0;
	xltoggle=XL_EXTENDED;
	if(lfile)free(lfile);
	lfile=strdup(line);
	break;
      }
      case 'd':{
        awaiting=0;
        delim=strdup(line);
        break;
      }
      case 't':{
        awaiting=0;
        if(lfile){//if we want to apply to another file
            FILE* l=input_file_open(lfile);
            if(l){
                interlace_to_file(line,l);
                fclose(l);
            }
            free(lfile);
            lfile=0;
        }
        else interlace_to_line(line);
        break;
      }
      case 'D':{
        awaiting=0;
        free(Fd);
        Fd=formatdup(line);
        break;
      }
      case 'F':{
        awaiting=0;
        free(Ff);
        Ff=formatdup(line);
        break;
      }
    }
    awaiting=0;
    if(tofree){
      free(tofree);
      tofree=0;
    }
    return;
  }

  //empty line?
  if(!*line)return;

  //ordinary line of code
  if(lfile){//applying to a file?
    FILE* l=input_file_open(lfile);
    if(l){
      if(xltoggle==XL_LINE){
	apply_to_file(line,l,false);
      }else{
	extended_to_file(line,l,false);
      }
      fclose(l);
    }
    free(lfile);
    lfile=0;
  }else{//just a line
    char* comment;
    array a;
    evaluate_line(line,&comment,true,a);
    if(comment){
      add_to_history(a);
    }
  }
}

int eval_control::format_splitline(char* line, vector<format_piece>& fmt){
  int countnums=0;
  char* c=line;

  format_piece p=(format_piece){'s',c,c};
  const char* delims=" \t,;:";
  while(*c && *c!='\n'){//iteration
    if(p.c=='d'){//number
      char* end;
      strtold(c,&end);//find the end of the number
      p.parsed_start=c;//this 'd' part of the condition only happens at the beginning of the number
      p.parsed_stop=end;
      if(end-c>0){
	fmt.push_back(p);//write down the number
	countnums++;
	p=(format_piece){'s',end,end};//new string
	c=end;
	continue;
      }else{
	p.c='s';
	c=end;
      }
    }else{//string
      if(isdigit(*c)||*c=='.'||*c=='-'){//found number, minus may start a number
	if(c==p.parsed_start){//a new number immediately, convert to number (this happens at the beginning)
	  p.c='d';
	  continue;
	}else{
	  if(strchr(delims,*(c-1)) ){//if previous is a delimiter
	    p.parsed_stop=c;
	    fmt.push_back(p);
	    p=(format_piece){'d',c,c};//new number
	    continue;
	  }else{//if not after delimiter, part of a string
	    //do nothing
	  }
	}
      }
    }
    c++;
  }
  //cleanup: if anything is still in the "p", add to format
  if(c>p.parsed_start){
    p.parsed_stop=c;
    fmt.push_back(p);
    if(p.c=='d')countnums++;
  }
  return countnums;
}

void eval_control::extended_to_file(char* l, FILE* f, bool file_or_line){
  vector<format_piece> fmt;

  //compile code
  list<code_chunk> code;
  int errcode=retrieve_code(l,code,file_or_line);
  if(errcode==2)return;

  int line=1;
  while(bufgets(f)){
    memory[memroot].clear();
    memory[memroot].push_back((itype)line);
    //parse. If no numbers, just return the same number and carry on
    fmt.clear();
    if(!format_splitline(buffer,fmt)){
      fputs(buffer,outstream);
      continue;
    }
    //here we fill the memory with slots from the input line
    memory[memdefault].clear();
    for(int i=0;i<fmt.size();i++){
      format_piece fp=fmt[i];
      if(fp.c=='d'){//number
	char end=*fp.parsed_stop;
	*fp.parsed_stop=0;//fake end of string
	memory[memdefault].push_back(value(fp.parsed_start));
	*fp.parsed_stop=end;//restore original character
      }
    }
    //evaluate the line
    array a;
    evaluate_code(code,false,a);
#ifndef NOSYSCALL
    if(shared::special_interrupt){
      if(shared::special_interrupt==INTERRUPT_BREAK){
	shared::special_interrupt=INTERRUPT_NONE;break;
      }
      if(shared::special_interrupt==INTERRUPT_CONT){
	shared::special_interrupt=INTERRUPT_NONE;continue;
      }
      return;
    }
#endif

    line++;
    if(!errcode){
      int outsize=a.size();
      int outpos=0;
      for(int i=0;i<fmt.size();i++){
	format_piece fp=fmt[i];
	if(fp.c=='d'){//number
	  if(outpos<outsize){
	    a[outpos].print(outstream,Fd,Ff);
	    outpos++;
	  }
	}else{//string
	  char end=*fp.parsed_stop;
	  *fp.parsed_stop=0;//fake end of string
	  fputs(fp.parsed_start,outstream);
	  *fp.parsed_stop=end;//restore original character
	}
      }
      //if there's anything left
      while(outpos<outsize){
	fputs(delim,outstream);
	a[outpos].print(outstream,Fd,Ff);
	outpos++;
      }
      putc('\n',outstream);//terminate line
    }
  }

}

void eval_control::add_to_history(array& arr){
  if(!toggle_history)return;
  histsize++;
  memory[-histsize]=arr;
}

eval_control::eval_control(){//initialization
#ifndef NOSYSCALL
  {
    tms tmbfr;
    start_time=times(&tmbfr);
  }
#endif
#ifdef TIMERAND
  srand(time(0));
#else
  {
    uint seed;
    FILE* f=fopen("/dev/urandom","r");
    fread(&seed,sizeof(uint),1,f);
    srand(seed);
    fclose(f);
  }
#endif
  funcount=1;
  functrl=0;
  maxit=65536;
  toggle_history=1;
  histsize=0;
  toggle_print=0;
  memroot=0;
  memdefault=1;
  mempos=0;
  awaiting=0;
  lfile=0;
  xltoggle=XL_LINE;

  outstream=stdout;

  delim=strdup(" ");
  Fd=formatdup("%d");
  Ff=formatdup("%g");
  bufsize=BUF_SIZE;
  buffer=(char*)malloc(bufsize);
}

eval_control::~eval_control(){
#if DEBUG>1
  char filenamebuf[64];
  for(t_fun::iterator i=funspace.begin();i!=funspace.end();++i){
    sprintf(filenamebuf,"xmltree_%03d.xml",i->first);
    FILE* f=fopen(filenamebuf,"w");
    i->second.treeplot(f);
    fclose(f);
  }
#endif
  if(outstream!=stdout)fclose(outstream);
  //free buffer
  free(buffer);
  //free other allocated strings
  free(delim);
  free(Fd);
  free(Ff);
}
